import numpy as np
from pxr import UsdGeom, Sdf
from functools import partial

import grill.usd as _usd

from ._qt import QtWidgets, QtGui


def _hsv_to_rgb(hsv):
    # taken directly from https://github.com/matplotlib/matplotlib/blob/f6e0ee49c598f59c6e6cf4eefe473e4dc634a58a/lib/matplotlib/colors.py#L1898-L1977
    # but don't want to depend on matplotlib for only it. Might change it the future.
    """Convert hsv values to rgb. All values assumed to be in range [0, 1]"""
    h, s, v = hsv[..., 0], hsv[..., 1], hsv[..., 2]
    r, g, b = np.empty_like(h), np.empty_like(h), np.empty_like(h)

    i = (h * 6.0).astype(int)
    f = (h * 6.0) - i
    p = v * (1.0 - s)
    q = v * (1.0 - s * f)
    t = v * (1.0 - s * (1.0 - f))

    idx = i % 6 == 0
    r[idx] = v[idx]
    g[idx] = t[idx]
    b[idx] = p[idx]

    idx = i == 1
    r[idx] = q[idx]
    g[idx] = v[idx]
    b[idx] = p[idx]

    idx = i == 2
    r[idx] = p[idx]
    g[idx] = v[idx]
    b[idx] = t[idx]

    idx = i == 3
    r[idx] = p[idx]
    g[idx] = q[idx]
    b[idx] = v[idx]

    idx = i == 4
    r[idx] = t[idx]
    g[idx] = p[idx]
    b[idx] = v[idx]

    idx = i == 5
    r[idx] = v[idx]
    g[idx] = p[idx]
    b[idx] = q[idx]

    idx = s == 0
    r[idx] = v[idx]
    g[idx] = v[idx]
    b[idx] = v[idx]
    rgb = np.stack([r, g, b], axis=-1)
    return rgb.reshape(hsv.shape)


def _random_colors(amount):
    return np.random.dirichlet(np.ones(3), size=amount)


def _color_spectrum(start: QtGui.QColor, end: QtGui.QColor, amount):
    *start_hsv, start_alpha = start.getHsvF()
    *end_hsv, end_alpha = end.getHsvF()
    return _hsv_to_rgb(np.linspace(start_hsv, end_hsv, amount))


class _DisplayColorEditor(QtWidgets.QFrame):
    # TODO: support alpha update
    # TODO: still experimental
    def __init__(self, primvar: UsdGeom.Primvar, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._primvar = primvar

        def setColorStyle(qobj, color):
            qobj.setStyleSheet(f"background-color: rgb{color.getRgb()[:3]}")  # don't want alpha

        def _pick_color(parent, launcher):
            result = QtWidgets.QColorDialog.getColor(launcher._color, parent, options=QtWidgets.QColorDialog.ShowAlphaChannel)
            if result.isValid():
                launcher._color = result
                setColorStyle(launcher, result)
                self._update_value()

        layout = QtWidgets.QVBoxLayout()
        options_layout = QtWidgets.QFormLayout()
        # If constant:
        #   Show single color option
        # Else:
        #   Show color range (start, finish) OR Random
        wavelength_layout = QtWidgets.QHBoxLayout()
        wavelength_options = QtWidgets.QButtonGroup()
        spectrum = QtWidgets.QRadioButton("Spectrum")
        spectrum.setChecked(True)
        self._random = random = QtWidgets.QRadioButton("Random")
        wavelength_options.addButton(spectrum)
        wavelength_options.addButton(random)

        for button in wavelength_options.buttons():
            wavelength_layout.addWidget(button)
        options_layout.addRow("Wavelength:", wavelength_layout)

        stackedWidget = QtWidgets.QStackedWidget()
        interpolation_options_box = QtWidgets.QComboBox()

        def _color_option_changed(*__):
            current_idx = int(bool(interpolation_options_box.currentIndex()))
            if random.isChecked():
                current_idx = 2
            stackedWidget.setCurrentIndex(current_idx)
            self._update_value()

        wavelength_options.buttonClicked.connect(_color_option_changed)

        if primvar and (value := primvar.Get()):
            start_color = QtGui.QColor.fromRgbF(*value[0])
            end_color = QtGui.QColor.fromRgbF(*value[-1])
        else:
            start_color = QtGui.QColor.fromHsv(359, 255, 255)
            end_color = QtGui.QColor.fromHsv(0, 255, 255)

        self._color_launchers = _color_launchers = {}
        for label, colors in (
                ("Color", (start_color, )),
                ("Range", (start_color, end_color)),  # GUI goes from 359 to 0 ):
        ):
            range_layout = QtWidgets.QHBoxLayout()
            range_layout.addStretch()
            _color_launchers[label] = []
            for color in colors:
                launcher = QtWidgets.QPushButton()
                _color_launchers[label].append(launcher)
                launcher._color = color
                setColorStyle(launcher, color)
                launcher.clicked.connect(partial(_pick_color, self, launcher))
                range_layout.addWidget(launcher)
            range_frame = QtWidgets.QFrame()
            range_layout.setContentsMargins(0, 0, 0, 0)
            range_frame.setLayout(range_layout)
            stackedWidget.addWidget(range_frame)

        random_layout = QtWidgets.QHBoxLayout()
        random_layout.addStretch()
        random_new = QtWidgets.QPushButton("New")
        random_new.clicked.connect(_color_option_changed)
        random_layout.addWidget(random_new)
        random_frame = QtWidgets.QFrame()
        random_layout.setContentsMargins(0, 0, 0, 0)
        random_frame.setLayout(random_layout)
        stackedWidget.addWidget(random_frame)
        for primvar_info in _usd._GeomPrimvarInfo:
            interpolation_options_box.addItem(primvar_info.interpolation())

        if primvar:
            interpolation_options_box.setCurrentText(primvar.GetInterpolation())
            stackedWidget.setCurrentIndex(int(bool(interpolation_options_box.currentIndex())))

        options_layout.addRow("Interpolation:", interpolation_options_box)
        layout.addLayout(options_layout)
        layout.addWidget(stackedWidget)
        self._interpolation_options_box = interpolation_options_box
        for button in wavelength_options.buttons():
            button.clicked.connect(_color_option_changed)
        interpolation_options_box.currentIndexChanged.connect(_color_option_changed)
        self.setLayout(layout)

    @property
    def _interpolation(self):
        return self._interpolation_options_box.currentText()

    @property
    def _size(self):
        if not self._primvar:
            return 1
        for i in _usd._GeomPrimvarInfo:
            if i.interpolation() == self._interpolation:
                return i.size(self._primvar.GetAttr().GetPrim())

    @property
    def _value(self):
        amount = self._size
        if self._random.isChecked():
            return _random_colors(amount)

        if self._interpolation == _usd._GeomPrimvarInfo.CONSTANT.interpolation():
            start = end = self._color_launchers["Color"][0]._color
        else:
            start, end = [pb._color for pb in self._color_launchers["Range"]]
        return _color_spectrum(start, end, amount)

    def _update_value(self, *__):
        if primvar := self._primvar:
            with Sdf.ChangeBlock():
                # Warning, when changing interpolation type in USDView, Hydra does not
                # seem to respect re-computation of the changed primvar for GlSl shader.
                # Workaround:
                # 1. Change draw mode to cards
                # 2. Change interpolation
                # 3. Change colors as desired!
                primvar.SetInterpolation(self._interpolation)
                primvar.SetElementSize(self._size)
                primvar.Set(self._value)



# Learnt from https://doc.qt.io/qtforpython-6/examples/example_external_networkx.html
from __future__ import annotations

import os
import math
import typing
import logging
import tempfile
import configparser
import networkx as nx
from itertools import chain
from functools import cache

from networkx import drawing

from . import _core
from ._qt import QtCore, QtGui, QtWidgets, QtSvg
_core._ensure_dot()

_logger = logging.getLogger(__name__)

_env_config = configparser.ConfigParser()
_env_config.read_dict(
    {
        "graph_view": {
            "via_svg": os.environ.get("GRILL_GRAPH_VIEW_VIA_SVG", 0),
            "svg_as_pixmap": os.environ.get("GRILL_SVG_VIEW_AS_PIXMAP", 0),
        }
    }
)
_GRAPHV_VIEW_VIA_SVG = _env_config.getboolean('graph_view', 'via_svg')
_USE_SVG_VIEWPORT = _env_config.getboolean('graph_view', 'svg_as_pixmap')

_IS_QT5 = QtCore.qVersion().startswith("5")

# TODO:
#   - Popup everytime a new graph is loaded in Houdini or Maya ( it's on _run_prog func line 1380 of agraph.py )
#       https://github.com/pygraphviz/pygraphviz/pull/514
#   - Should toggling "precise source layer" on LayerStack compostiion view preserve node position for _GraphViewer?
#   - Tooltip on nodes for _GraphViewer
#   - Context menu items
#   - Ability to move further in canvas after Nodes don't exist
#   - when switching a node left to right with precise source layers, the source node plugs do not refresh if we're moving the target node
#   - refactor conditionals for _GraphSVGViewer from the description module


_NO_PEN = QtGui.QPen(QtCore.Qt.NoPen)

_DOT_ENVIRONMENT_ERROR = """In order to display composition arcs in a graph,
the 'dot' command must be available on the current environment.

Please make sure graphviz is installed and 'dot' available on the system's PATH environment variable.

For more details on installing graphviz, visit https://pygraphviz.github.io/documentation/stable/install.html
"""


def _convert_graphviz_to_html_label(label):
    # TODO: these checks below rely on internals from the grill (layer stack composition uses record shapes, connection viewer uses html)
    if label.startswith("{"):  # We're a record. Split the label into individual fields
        fields = label.strip("{}").split("|")
        label = '<table>'
        for index, field in enumerate(fields):
            port, text = field.strip("<>").split(">", 1)
            bgcolor = "white" if index % 2 == 0 else "#f0f6ff"  # light blue
            # text = f'<font color="#3e4444">{text}</font>'
            text = f'<font color="#242828">{text}</font>'
            label += f"<tr><td port='{port}' bgcolor='{bgcolor}'>{text}</td></tr>"
        label += "</table>"
    elif label.startswith("<"):
        # Contract: HTML graphviz labels start with a double <<, additionally, ROUNDED is internal to graphviz
        # QGraphicsTextItem seems to have trouble with HTML rounding, so we're controlling this via paint + custom style
        label = label.removeprefix("<").removesuffix(">").replace('table border="1" cellspacing="2" style="ROUNDED"', "table")
    return label


@cache
def _dot_2_svg(sourcepath):
    print(f"Creating svg for: {sourcepath}")
    targetpath = f"{sourcepath}.svg"
    args = [_core._which("dot"), sourcepath, "-Tsvg", "-o", targetpath]
    error, __ = _core._run(args)
    return error, targetpath


class _Node(QtWidgets.QGraphicsTextItem):

    def __init__(self, parent=None, label="", color="", fillcolor="", plugs: tuple =None, active_plugs: set = frozenset(), visible=True):
        super().__init__(parent)
        self._edges = []
        self._plugs = plugs = dict(zip(plugs, range(len(plugs)))) or {}  # {identifier: index}

        plug_items = {}  # {index: (QEllipse, QEllipse)}
        radius = 4
        def _plug_item():
            item = QtWidgets.QGraphicsEllipseItem(-radius, -radius, 2 * radius, 2 * radius)
            item.setPen(_NO_PEN)
            return item
        self._active_plugs = active_plugs
        self._active_plugs_by_side = dict()  # {index: {left[int]: {}, right[int]: {}}
        for plug_index in active_plugs:
            plug_items[plugs[plug_index]] = (_plug_item(), _plug_item())
            self._active_plugs_by_side[plugs[plug_index]] = {0: dict(), 1: dict()}
        self._plug_items = plug_items
        self._pen = QtGui.QPen(QtGui.QColor(color), 1, QtCore.Qt.SolidLine, QtCore.Qt.RoundCap, QtCore.Qt.RoundJoin)
        self._fillcolor = QtGui.QColor(fillcolor)
        self.setHtml("<style>th, td {text-align: center;padding: 3px}</style>" + _convert_graphviz_to_html_label(label))
        # Temp measure: allow PySide6 interaction, but not in PySide2 as this causes a crash on windows:
        # https://stackoverflow.com/questions/67264846/pyqt5-program-crashes-when-editable-qgraphicstextitem-is-clicked-with-right-mo
        # https://bugreports.qt.io/browse/QTBUG-89563
        self._default_text_interaction = QtCore.Qt.LinksAccessibleByMouse if _IS_QT5 else QtCore.Qt.TextBrowserInteraction
        if visible:
            self.setTextInteractionFlags(self._default_text_interaction)
            self.setFlag(QtWidgets.QGraphicsItem.ItemIsMovable)
            self.setFlag(QtWidgets.QGraphicsItem.ItemIsSelectable)
            self.setAcceptHoverEvents(True)
        else:
            self.setVisible(False)
        self.setFlag(QtWidgets.QGraphicsItem.ItemSendsGeometryChanges)
        self.setCacheMode(QtWidgets.QGraphicsItem.DeviceCoordinateCache)

    def _overrideCursor(self, event):
        if event.modifiers() == QtCore.Qt.ControlModifier:
            self.setCursor(QtGui.Qt.PointingHandCursor)
        elif event.modifiers() == QtCore.Qt.AltModifier:
            self.setCursor(QtGui.Qt.ClosedHandCursor)
            self.setTextInteractionFlags(QtCore.Qt.NoTextInteraction)
        else:
            self.setTextInteractionFlags(self._default_text_interaction)
            self.setCursor(QtGui.Qt.ArrowCursor)

    def hoverMoveEvent(self, event: QtWidgets.QGraphicsSceneHoverEvent) -> None:
        self._overrideCursor(event)
        super().hoverMoveEvent(event)

    def mouseReleaseEvent(self, event):
        if event.button() == QtCore.Qt.LeftButton and event.modifiers() == QtCore.Qt.ControlModifier:
            self.linkActivated.emit("")
        else:
            super().mouseReleaseEvent(event)

    def hoverEnterEvent(self, event):
        self._overrideCursor(event)
        super().hoverEnterEvent(event)

    def hoverLeaveEvent(self, event):
        self.setCursor(QtGui.Qt.ArrowCursor)
        self.setTextInteractionFlags(self._default_text_interaction)
        super().hoverLeaveEvent(event)

    def paint(self, painter: QtGui.QPainter, option: QtWidgets.QStyleOptionGraphicsItem, widget: QtWidgets.QWidget) -> None:
        painter.setRenderHints(QtGui.QPainter.Antialiasing)
        painter.setPen(self._pen)
        rect_path = QtGui.QPainterPath()
        round_args = self.boundingRect().adjusted(1, 1, -1, -1), 6, 6
        rect_path.addRoundedRect(*round_args)
        painter.fillPath(rect_path, self._fillcolor)
        painter.drawRoundedRect(*round_args)
        return super().paint(painter, option, widget)

    def add_edge(self, edge: _Edge):
        self._edges.append(edge)

    def itemChange(self, change: QtWidgets.QGraphicsItem.GraphicsItemChange, value):
        if change == QtWidgets.QGraphicsItem.ItemPositionHasChanged:
            for edge in self._edges:
                edge.adjust()
        return super().itemChange(change, value)

    def _activatePlug(self, edge, plug_index, side, position):
        if plug_index is None:
            return  # we're at the center, nothing to draw nor activate
        plugs_by_side = self._active_plugs_by_side[plug_index]  # {index: {left[int]: {}, right[int]: {}}
        plugs_by_side[side][edge] = True
        other_side = bool(not side)
        inactive_plugs = plugs_by_side[other_side]
        inactive_plugs.pop(edge, None)
        plug_items = self._plug_items[plug_index]  # {index: (QEllipse, QEllipse)}
        if not inactive_plugs:
            plug_items[other_side].setVisible(False)
        this_item = plug_items[side]
        this_item.setVisible(True)
        this_item.setBrush(edge._brush)
        plug_items[side].setPos(position)


class _Edge(QtWidgets.QGraphicsItem):
    def __init__(self, source: _Node, target: _Node, *, source_plug: int =None, target_plug: int =None, label="", color="", is_bidirectional=False, parent: QtWidgets.QGraphicsItem = None):
        super().__init__(parent)
        source.add_edge(self)
        target.add_edge(self)
        self._source = source
        self._target = target
        self._source_plug = source_plug
        self._target_plug = target_plug
        self._is_source_plugged = source_plug is not None
        self._is_target_plugged = target_plug is not None
        self._is_cycle = is_cycle = source == target

        self._plug_positions = plug_positions = {}
        outer_shift = 10  # surrounding rect has ~5 px top and bottom

        for node, plug, max_plug_idx in (source, source_plug, max(source._plugs.values(), default=0)), (target, target_plug, max(target._plugs.values(), default=0)):
            bounds = node.boundingRect()
            if plug is None:
                plug_positions[node, plug] = {None: QtCore.QPointF(bounds.right() - 5, bounds.height() / 2 - 20) if is_cycle else bounds.center()}
                continue
            # max_plug_idx can be 0, so we add 1 since this needs to be 1-index based
            port_size = (bounds.height() - outer_shift) / (max_plug_idx + 1)
            y_pos = (plug * port_size) + (port_size / 2) + (outer_shift / 2)
            plug_positions[node, plug] = {
                0: QtCore.QPointF(0, y_pos),  # left
                1: QtCore.QPointF(bounds.right(), y_pos),  # right
            }

        self._width = 1.5
        self._arrow_size = 15
        self._bidirectional_shift = 20 if is_bidirectional else 0
        self._line = QtCore.QLineF()
        self.setZValue(-1)

        self._spline_path = QtGui.QPainterPath() if (self._is_source_plugged or self._is_target_plugged) else None

        self._colors = colors = color.split(":")
        main_color = QtGui.QColor(colors[0])
        if label:
            self._label_text = QtWidgets.QGraphicsTextItem(label, self)
            self._label_text.setDefaultTextColor(main_color)
            self._label_text.setHtml(f"<b>{label}</b>")
        else:
            self._label_text = None
        self._pen = QtGui.QPen(main_color, self._width, QtCore.Qt.SolidLine, QtCore.Qt.RoundCap, QtCore.Qt.RoundJoin)
        self._brush = QtGui.QBrush(main_color)
        if _IS_QT5:
            self._brush.setStyle(QtGui.Qt.SolidPattern)
            self._brush.setColor(main_color)

        self.adjust()

    def boundingRect(self) -> QtCore.QRectF:
        """Override from QtWidgets.QGraphicsItem

        Returns:
            QRect: Return node bounding rect
        """
        if self._is_cycle:
            top_left = self._cycle_start_position
            bottom_right = top_left + QtCore.QPointF(50, 50)
        else:
            top_left, bottom_right = self._line.p1(), self._line.p2()

        width, arrow_size = self._width, max(self._arrow_size, 50)
        top_shift, bottom_shift = -width - arrow_size, width + arrow_size
        return QtCore.QRectF(top_left, bottom_right).normalized().adjusted(top_shift, top_shift, bottom_shift, bottom_shift)

    @property
    def _cycle_start_position(self):
        if not self._is_source_plugged:
            return self._source.pos() + self._plug_positions[self._source, self._source_plug][None]

        return self._line.p1() + QtCore.QPointF(-3, -31)

    def adjust(self):
        """Update edge position from source and target node following Node::itemChange."""
        self.prepareGeometryChange()
        source_pos = self._source.pos()
        target_pos = self._target.pos()
        target_bounds = self._target.boundingRect()

        source_on_left = self._is_cycle or (self._source.boundingRect().center().x() + source_pos.x() < target_bounds.center().x() + target_pos.x())

        is_source_plugged = self._is_source_plugged
        is_target_plugged = self._is_target_plugged
        source_side = source_on_left if is_source_plugged else None
        target_side = not source_side if is_target_plugged else None
        source_point = source_pos + self._plug_positions[self._source, self._source_plug][source_side]
        target_point = target_pos + self._plug_positions[self._target, self._target_plug][target_side]

        if not is_target_plugged:
            line = QtCore.QLineF(source_point, target_point)
            if not self._spline_path and self._bidirectional_shift and source_point != target_point:
                # offset in case of bidirectional connections when we are not using splines (as lines would overlap)
                line = _parallel_line(line, distance=self._bidirectional_shift, head_offset=0)

            # Check if there is an intersection on the target node to know where to draw the arrow
            if _IS_QT5:
                intersect_method = line.intersect
                bounded_intersection = QtCore.QLineF.IntersectType.BoundedIntersection
            else:
                intersect_method = line.intersects
                bounded_intersection = QtCore.QLineF.IntersectionType.BoundedIntersection

            for each in (
                    QtCore.QLineF((topLeft:=target_bounds.topLeft()) + target_pos, (topRight:=target_bounds.topRight()) + target_pos),  # top
                    QtCore.QLineF(topLeft + target_pos, (bottomLeft:=target_bounds.bottomLeft()) + target_pos),  # left
                    QtCore.QLineF(bottomLeft + target_pos, (bottomRight:=target_bounds.bottomRight()) + target_pos),  # bottom
                    QtCore.QLineF(bottomRight + target_pos, topRight + target_pos),  # right
            ):  # TODO: how to make this more efficient?
                intersection, intersection_point = intersect_method(each)
                if intersection == bounded_intersection:
                    target_point = intersection_point
                    break
            else:
                target_point = line.p2()

        self._line = line = QtCore.QLineF(source_point, target_point)
        if self._spline_path:
            length = line.length()
            falloff = (length / 100) ** 2 if length < 100 else 1
            control_point_shift = (1 if source_on_left else -1) * 75 * falloff

            control_point1 = source_point + QtCore.QPointF(control_point_shift, 0) if is_source_plugged else source_point
            control_point2 = target_point + QtCore.QPointF(-control_point_shift, 0) if is_target_plugged else target_point

            self._spline_path = QtGui.QPainterPath()
            self._spline_path.moveTo(source_point)
            self._spline_path.cubicTo(control_point1, control_point2, target_point)

        self._source._activatePlug(self, self._source_plug, source_side, source_point)
        self._target._activatePlug(self, self._target_plug, target_side, target_point)
        if self._label_text:
            self._label_text.setPos((source_point + target_point) / 2)

    def paint(self, painter: QtGui.QPainter, option: QtWidgets.QStyleOptionGraphicsItem, widget=None):
        """Draw Edge following ``Edge.adjust(...)``"""
        painter.setRenderHints(QtGui.QPainter.Antialiasing)
        painter.setPen(self._pen)
        if self._is_cycle:
            self._paint_cyclic_arrow(painter, self._cycle_start_position)
        else:
            total_colors = enumerate(self._colors)
            __, main_color = next(total_colors)  # draw first without offset and with current color

            if self._spline_path:
                arrow_head_end_point = self._spline_path.pointAtPercent(1)
                arrow_head_start_point = self._spline_path.pointAtPercent(.95)
                parallel_paths = []
                for i, color in reversed(list(total_colors)):
                    stroker = QtGui.QPainterPathStroker()
                    stroker.setCapStyle(QtCore.Qt.PenCapStyle.RoundCap)
                    stroker.setWidth(i*5)
                    parallel_paths.append((color, stroker.createStroke(self._spline_path)))

                painter.drawPath(self._spline_path)
                for color, parallel_path in parallel_paths:
                    self._pen.setColor(color)
                    painter.setPen(self._pen)  # Set the color and thickness
                    painter.drawPath(parallel_path)
            else:  # painting as a straight line
                arrow_head_start_point = self._line.p1()
                arrow_head_end_point = self._line.p2()

                painter.drawLine(self._line)
                for index, color in total_colors:
                    self._pen.setColor(color)
                    shift = int((index+1)/2) * 1.5 * 3
                    painter.setPen(self._pen)
                    painter.drawLine(_parallel_line(self._line, shift if index % 2 == 0 else -shift, head_offset=11))

            self._pen.setColor(main_color)
            painter.setPen(self._pen)
            self._draw_arrow_head(painter, arrow_head_start_point, arrow_head_end_point)

    def _paint_cyclic_arrow(self, painter: QtGui.QPainter, source_pos: QtCore.QPointF):
        center_x, center_y = source_pos.toTuple()

        for index, color in enumerate(self._colors):
            self._pen.setColor(color)
            painter.setPen(self._pen)
            arc_offset = index * 1.5
            box_y = center_y + (1.5 * index)
            box_size = (20 - arc_offset) * 2
            start_angle = 135 - (5 * index)
            finish_angle = -270 + arc_offset
            arrow_path = QtGui.QPainterPath()
            arrow_path.arcMoveTo(center_x, box_y, box_size, box_size, start_angle)
            arrow_path.arcTo(center_x, box_y, box_size, box_size, start_angle, finish_angle)
            painter.drawPath(arrow_path)

        self._pen.setColor(self._colors[0])
        painter.setPen(self._pen)
        start = QtCore.QPointF(center_x+7, center_y+6)
        end = QtCore.QPointF(center_x+4, center_y+9)
        self._draw_arrow_head(painter, start, end)

    def _draw_arrow_head(self, painter: QtGui.QPainter, start: QtCore.QPointF, end: QtCore.QPointF):
        """Draw arrow from start point to end point."""

        def point_in_direction(start: QtCore.QPointF, end: QtCore.QPointF, distance: float) -> QtCore.QPointF:
            direction = QtGui.QVector2D(end - start)
            direction.normalize()
            return start + direction.toPointF() * distance

        brush = self._brush
        painter.setBrush(brush)

        line = QtCore.QLineF(end, start)

        head_tilt = math.pi / 3
        arrow_size = self._arrow_size
        angle = math.atan2(-line.dy(), line.dx())
        left_tilt, right_tilt = angle + head_tilt, angle + math.pi - head_tilt
        arrow_p1 = line.p1() + QtCore.QPointF(math.sin(left_tilt) * arrow_size, math.cos(left_tilt) * arrow_size)
        arrow_p2 = line.p1() + QtCore.QPointF(math.sin(right_tilt) * arrow_size, math.cos(right_tilt) * arrow_size)

        arrow_head = QtGui.QPolygonF()
        arrow_head.append(line.p1())
        arrow_head.append(arrow_p1)
        arrow_head.append(point_in_direction(end, start, 10))
        arrow_head.append(arrow_p2)
        painter.drawPolygon(arrow_head)


def _parallel_line(line, distance, head_offset=0):
    direction = line.unitVector()
    # Calculate the perpendicular vector by rotating the direction vector by 90 degrees, then the offset
    perpendicular = QtCore.QPointF(-direction.dy(), direction.dx()) * (distance / 2)
    parallel_line = QtCore.QLineF(line.p1() + perpendicular, line.p2() + perpendicular)
    if head_offset:
        parallel_line.setLength(parallel_line.length() - head_offset)
    return parallel_line


_EVENT_POSITION_FUNC = QtGui.QMouseEvent.globalPos if _IS_QT5 else lambda event: event.globalPosition().toPoint()


class _GraphicsViewport(QtWidgets.QGraphicsView):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._dragging = False
        self._last_pan_pos = QtCore.QPoint()
        self._rubber_band = QtWidgets.QRubberBand(QtWidgets.QRubberBand.Rectangle, self)
        self._start_rubber_band_pos = QtCore.QPoint()

    def wheelEvent(self, event):
        modifiers = event.modifiers()

        if modifiers == QtCore.Qt.ControlModifier:
            zoom_factor = 1.2 if event.angleDelta().y() > 0 else 0.8
            self.scale(zoom_factor, zoom_factor)
        elif modifiers == QtCore.Qt.AltModifier:
            self.horizontal_pan(event)
        else:
            # Pan vertically when no modifier key is pressed
            self.vertical_pan(event)

    def mousePressEvent(self, event):
        if event.button() == QtCore.Qt.MiddleButton:
            self._dragging = True
            QtWidgets.QApplication.setOverrideCursor(QtGui.Qt.ClosedHandCursor)
            self._last_pan_pos = _EVENT_POSITION_FUNC(event)
            event.accept()
        elif event.button() == QtCore.Qt.LeftButton and event.modifiers() == QtCore.Qt.NoModifier:
            self._start_rubber_band_pos = event.pos()
            self._rubber_band.setGeometry(QtCore.QRect(self._start_rubber_band_pos, QtCore.QSize()))
            self._rubber_band.show()

        return super().mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        if event.button() == QtCore.Qt.MiddleButton:
            self._dragging = False
            QtWidgets.QApplication.restoreOverrideCursor()
            event.accept()
        elif event.button() == QtCore.Qt.LeftButton and event.modifiers() == QtCore.Qt.NoModifier:
            self._rubber_band.hide()
            for item in self._get_items_in_rubber_band():
                item.setSelected(True)
        return super().mouseReleaseEvent(event)

    def _get_items_in_rubber_band(self):
        rubber_band_rect = self._rubber_band.geometry()
        scene_rect = self.mapToScene(rubber_band_rect).boundingRect()
        return self.scene().items(scene_rect)

    def mouseMoveEvent(self, event):
        if self._dragging and event.buttons() == QtCore.Qt.MiddleButton:
            # Pan the scene when middle mouse button is held down
            delta = _EVENT_POSITION_FUNC(event) - self._last_pan_pos
            self.horizontalScrollBar().setValue(self.horizontalScrollBar().value() - delta.x())
            self.verticalScrollBar().setValue(self.verticalScrollBar().value() - delta.y())
            self._last_pan_pos = _EVENT_POSITION_FUNC(event)
            return
        elif event.buttons() == QtCore.Qt.LeftButton and event.modifiers() == QtCore.Qt.NoModifier:
            self._rubber_band.setGeometry(QtCore.QRect(self._start_rubber_band_pos, event.pos()).normalized())
        return super().mouseMoveEvent(event)

    def horizontal_pan(self, event):
        delta = event.angleDelta().x()
        scroll_bar = self.horizontalScrollBar()
        scroll_bar.setValue(scroll_bar.value() - delta)

    def vertical_pan(self, event):
        delta = event.angleDelta().y()
        scroll_bar = self.verticalScrollBar()
        scroll_bar.setValue(scroll_bar.value() - delta)

    def keyPressEvent(self, event):
        if event.key() == QtCore.Qt.Key_F:
            self._focus_view_requested()
        return super().keyPressEvent(event)

    def _focus_view_requested(self):
        if selected_items := self.scene().selectedItems():
            bounding_rect = QtCore.QRectF()
            for item in selected_items:
                bounding_rect |= item.sceneBoundingRect()
        else:  # if no items have been selected, focus on all items
            bounding_rect = self.scene().itemsBoundingRect()

        self.fitInView(bounding_rect, QtCore.Qt.KeepAspectRatio)


class GraphView(_GraphicsViewport):
    def __init__(self, graph: nx.DiGraph = None, parent=None):
        super().__init__(parent=parent)
        self._filter_edges = None
        self._graph = graph
        self._scene = QtWidgets.QGraphicsScene()
        self.setScene(self._scene)

        self._nodes_map = {}  # {str: Node}

        self._load_graph(None)

        self.sticky_nodes = list()
        self._viewing = set()
        self.url_id_prefix = ""

    def _graph_url_changed(self, *_, **__):
        sender = self.sender()
        key = next((k for k, v in self._nodes_map.items() if v==sender), None)
        if key is None:
            raise LookupError(f"Could not find sender {sender} in nodes map")

        self.view((key,))

    def view(self, node_indices: tuple):
        self._viewing = frozenset(node_indices)
        graph = self._graph
        if not graph:
            return
        successors = chain.from_iterable(graph.successors(index) for index in node_indices)
        predecessors = chain.from_iterable(graph.predecessors(index) for index in node_indices)
        nodes_of_interest = chain(self.sticky_nodes, node_indices, successors, predecessors)
        subgraph = graph.subgraph(nodes_of_interest)

        filters = {}
        if self.filter_edges:
            filters['filter_edge'] = self.filter_edges
        if filters:
            subgraph = nx.subgraph_view(subgraph, **filters)

        self._load_graph(subgraph)

    @property
    def graph(self):
        return self._graph

    @property
    def filter_edges(self):
        return self._filter_edges

    @filter_edges.setter
    def filter_edges(self, value):
        if value == self._filter_edges:
            return
        # self._subgraph_dot_path.cache_clear()
        if value:
            predicate = lambda *edge: value(*edge) or bool({edge[0], edge[1]}.intersection(self.sticky_nodes))
        else:
            predicate = None
        self._filter_edges = predicate

    @graph.setter
    def graph(self, graph):
        self._graph = graph
        self._load_graph(graph)

    def _load_graph(self, graph):
        if not graph:
            return
        self.scene().clear()
        self.viewport().update()

        if not _core._which("dot"):  # dot has not been installed
            print(_DOT_ENVIRONMENT_ERROR)
            text_item = QtWidgets.QGraphicsTextItem()
            text_item.setPlainText(_DOT_ENVIRONMENT_ERROR)
            text_item.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse if _IS_QT5 else QtCore.Qt.TextBrowserInteraction)
            self.scene().addItem(text_item)
            return

        try:  # exit early if pygraphviz is not installed, needed for positions
            positions = drawing.nx_agraph.graphviz_layout(graph, prog='dot')
        except ImportError as exc:
            message = str(exc)
            print(message)
            text_item = QtWidgets.QGraphicsTextItem()
            text_item.setPlainText(message)
            self.scene().addItem(text_item)
            return

        print("LOADING GRAPH")
        self._nodes_map.clear()
        edge_color = graph.graph.get('edge', {}).get("color", "")

        def _add_node(nx_node):
            node_data = graph.nodes[nx_node]
            item = _Node(
                label=node_data.get('label', str(nx_node)),
                color=graph.graph.get('node', {}).get("color", ""),
                fillcolor=graph.graph.get('node', {}).get("fillcolor", "white"),
                plugs=node_data.get('plugs', {}),
                visible=node_data.get('style', "") != "invis",
                active_plugs=node_data.get('active_plugs', set()),
            )
            item.linkActivated.connect(self._graph_url_changed)
            self.scene().addItem(item)
            for each_plug in chain.from_iterable(item._plug_items.values()):
                self.scene().addItem(each_plug)
            return item

        max_y = max(pos[1] for pos in positions.values())
        for nx_node in graph:
            self._nodes_map[nx_node] = node = _add_node(nx_node)
            x_pos, y_pos = positions[nx_node]
            # SVG and dot have inverted coordinates, let's flip Y
            y_pos = max_y - y_pos
            bounds = node.boundingRect()
            # x, y refer to the node's center. Calculate the node position now (top left corner is 0,0)
            node.setPos(x_pos - bounds.width() / 2, y_pos - bounds.height() / 2)

        if isinstance(graph, nx.MultiDiGraph):
            edges = graph.edges
            edge_data_getter = lambda source, target, port: graph.edges[source, target, port]
        else:
            edges = ((source, target, None) for source, target in graph.edges)
            edge_data_getter = lambda source, target, port: graph.edges[source, target]

        for source_id, target_id, port in edges:
            source = self._nodes_map[source_id]
            target = self._nodes_map[target_id]
            is_bidirectional = graph.has_edge(target_id, source_id)
            edge_data = edge_data_getter(source_id, target_id, port)
            color = edge_data.get('color', edge_color)
            label = edge_data.get('label', '')
            kwargs = dict()
            if source._plugs or target._plugs:
                kwargs['target_plug'] = target._plugs[edge_data['headport']] if edge_data.get('headport') is not None else None
                kwargs['source_plug'] = source._plugs[edge_data['tailport']] if edge_data.get('tailport') is not None else None
            edge = _Edge(source, target, color=color, label=label, is_bidirectional=is_bidirectional, **kwargs)
            self.scene().addItem(edge)


class _Dot2SvgSignals(QtCore.QObject):
    error = QtCore.Signal(str)
    result = QtCore.Signal(str)


class _Dot2Svg(QtCore.QRunnable):
    def __init__(self, source_fp, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.signals = _Dot2SvgSignals()
        self.source_fp = source_fp

    @QtCore.Slot()
    def run(self):
        if not _core._which("dot"):
            self.signals.error.emit(_DOT_ENVIRONMENT_ERROR)
            return
        error, svg_fp = _dot_2_svg(self.source_fp)
        self.signals.error.emit(error) if error else self.signals.result.emit(svg_fp)


class _SvgPixmapViewport(_GraphicsViewport):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        scene = QtWidgets.QGraphicsScene(self)
        self.setScene(scene)

    def load(self, filepath):
        scene = self.scene()
        scene.clear()

        renderer = QtSvg.QSvgRenderer(filepath)
        image = QtGui.QImage(renderer.defaultSize() * 1.5, QtGui.QImage.Format_ARGB32)
        image.fill(QtCore.Qt.transparent)

        painter = QtGui.QPainter(image)
        renderer.render(painter)
        painter.end()

        pixmap = QtGui.QPixmap.fromImage(image)
        self._svg_item = QtWidgets.QGraphicsPixmapItem(pixmap)
        scene.addItem(self._svg_item)


class _DotViewer(QtWidgets.QFrame):
    _svg_viewport_placeholder_signal = QtCore.Signal(object)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        layout = QtWidgets.QVBoxLayout()
        # After some experiments, QWebEngineView brings nicer UX and speed than QGraphicsSvgItem and QSvgWidget
        if not _USE_SVG_VIEWPORT:
            if QtWidgets.__package__ == "PySide6":
                # PySide-6.6.0 and 6.6.1 freeze when QtWebEngineWidgets is imported in Python-3.12 so inlining here until fixed
                # Newest working combination for me in Windows is 3.11 + PySide-6.5.3 as of Feb 3rd 2024
                from PySide6 import QtWebEngineWidgets
            else:
                from PySide2 import QtWebEngineWidgets
            self._graph_view = QtWebEngineWidgets.QWebEngineView(parent=self)
            self.urlChanged = self._graph_view.urlChanged
        else:
            self.urlChanged = self._svg_viewport_placeholder_signal
            self._graph_view = _SvgPixmapViewport(parent=self)

        self._error_view = QtWidgets.QTextBrowser(parent=self)
        layout.addWidget(self._graph_view)
        layout.addWidget(self._error_view)
        layout.setContentsMargins(0, 0, 0, 0)
        self._error_view.setVisible(False)
        self.setLayout(layout)
        self._dot2svg = None
        self._threadpool = QtCore.QThreadPool()
        if not _USE_SVG_VIEWPORT:
            # otherwise it seems invisible
            self.setMinimumHeight(100)

    def setDotPath(self, path):
        if self._dot2svg:  # forget about previous, unfinished runners
            self._dot2svg.signals.error.disconnect()
            self._dot2svg.signals.result.disconnect()

        self._dot2svg = dot2svg = _Dot2Svg(path)
        dot2svg.signals.error.connect(self._on_dot_error)
        dot2svg.signals.result.connect(self._on_dot_result)
        self._threadpool.start(dot2svg)

    def _on_dot_error(self, message):
        self._error_view.setVisible(True)
        self._graph_view.setVisible(False)
        self._error_view.setText(message)

    def _on_dot_result(self, filepath):
        self._error_view.setVisible(False)
        self._graph_view.setVisible(True)
        if not _USE_SVG_VIEWPORT:
            filepath = QtCore.QUrl.fromLocalFile(filepath)
        self._graph_view.load(filepath)


class _GraphSVGViewer(_DotViewer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.urlChanged.connect(self._graph_url_changed)
        self.sticky_nodes = list()
        self._graph = None
        self._viewing = frozenset()
        self._filter_edges = None

    @property
    def filter_edges(self):
        return self._filter_edges

    @filter_edges.setter
    def filter_edges(self, value):
        if value == self._filter_edges:
            return
        self._subgraph_dot_path.cache_clear()
        if value:
            predicate = lambda *edge: value(*edge) or bool({edge[0], edge[1]}.intersection(self.sticky_nodes))
        else:
            predicate = None
        self._filter_edges = predicate

    @property
    def url_id_prefix(self):
        return "_node_id_"

    def _graph_url_changed(self, url: QtCore.QUrl):
        node_uri = url.toString()
        node_uri_stem = node_uri.split("/")[-1]
        if node_uri_stem.startswith(self.url_id_prefix):
            index = node_uri_stem.split(self.url_id_prefix)[-1]
            self.view([int(index)] if index.isdigit() else [index])

    @cache
    def _subgraph_dot_path(self, node_indices: tuple):
        graph = self.graph
        if not graph:
            raise RuntimeError(f"'graph' attribute not set yet on {self}. Can't view nodes {node_indices}")
        successors = chain.from_iterable(graph.successors(index) for index in node_indices)
        predecessors = chain.from_iterable(graph.predecessors(index) for index in node_indices)
        nodes_of_interest = chain(self.sticky_nodes, node_indices, successors, predecessors)
        subgraph = graph.subgraph(nodes_of_interest)

        filters = {}
        if self.filter_edges:
            print(f"{self.filter_edges=}")
            filters['filter_edge'] = self.filter_edges
        if filters:
            subgraph = nx.subgraph_view(subgraph, **filters)

        fd, fp = tempfile.mkstemp()
        try:
            nx.nx_agraph.write_dot(subgraph, fp)
        except ImportError as exc:
            error = f"{exc}\n\n{_DOT_ENVIRONMENT_ERROR}"
        else:
            error = ""
        return error, fp

    def view(self, node_indices: typing.Iterable):
        error, dot_path = self._subgraph_dot_path(tuple(node_indices))
        self._viewing = frozenset(node_indices)
        if error:
            self._on_dot_error(error)
        else:
            self.setDotPath(dot_path)

    @property
    def graph(self):
        return self._graph

    @graph.setter
    def graph(self, graph):
        self._subgraph_dot_path.cache_clear()
        self.sticky_nodes.clear()
        self._graph = graph


if _GRAPHV_VIEW_VIA_SVG:
    _GraphViewer = _GraphSVGViewer
else:
    _GraphViewer = GraphView



"""Inspecting, authoring and editing foundational tools for the pipeline.

.. data:: Repository

    :class:`contextvars.ContextVar` for the global asset repository location.

    It's value must always be set to a :class:`pathlib.Path`.

    .. attention::
        By default, no value has been set. Ensure to set it before performing any creation operation.

    Example:
        >>> Repository.get()  # not set
        Traceback (most recent call last):
          File "<input>", line 1, in <module>
        LookupError: <ContextVar name='Repository' at 0x00000207F0A12B88>
        >>> import tempfile
        >>> from pathlib import Path
        >>> Repository.set(Path(tempfile.mkdtemp()))
        <Token var=<ContextVar name='Repository' at 0x00000213A46FF900> at 0x00000213C6A9F0C0>
        >>> Repository.get()
        WindowsPath('C:/Users/CHRIST~1/AppData/Local/Temp/tmp767wqaya')

"""
from __future__ import annotations

import types
import typing
import logging
import functools
import itertools
import contextlib
import contextvars
from pathlib import Path
from pprint import pformat

import networkx as nx
from pxr import UsdGeom, Usd, Sdf, Kind, Ar

from grill.tokens import ids
from grill.names import UsdAsset

from .. import usd as _usd

logger = logging.getLogger(__name__)

Repository = contextvars.ContextVar('Repository')

_TAXA_KEY = 'taxa'
_FIELDS_KEY = 'fields'
_ASSETINFO_KEY = 'grill'
_ASSETINFO_TAXA_KEY = f'{_ASSETINFO_KEY}:{_TAXA_KEY}'
_ASSETINFO_FIELDS_KEY = f'{_ASSETINFO_KEY}:{_FIELDS_KEY}'

# Taxonomy rank handles the grill classification and grouping of assets.
_TAXONOMY_NAME = 'Taxonomy'
_TAXONOMY_ROOT_PATH = Sdf.Path.absoluteRootPath.AppendChild(_TAXONOMY_NAME)
_TAXONOMY_UNIQUE_ID = ids.CGAsset.cluster  # High level organization of our assets.
_TAXONOMY_FIELDS = types.MappingProxyType({_TAXONOMY_UNIQUE_ID.name: _TAXONOMY_NAME})
_ASSETINFO_TAXON_KEY = f'{_ASSETINFO_FIELDS_KEY}:{_TAXONOMY_UNIQUE_ID.name}'

_CATALOGUE_NAME = 'Catalogue'
_CATALOGUE_ROOT_PATH = Sdf.Path.absoluteRootPath.AppendChild(_CATALOGUE_NAME)
_CATALOGUE_ID = ids.CGAsset.kingdom  # where all existing units will be "discoverable"
_CATALOGUE_FIELDS = types.MappingProxyType({_CATALOGUE_ID.name: _CATALOGUE_NAME})

_INHERITED_ROOT_PATH = Sdf.Path.absoluteRootPath.AppendChild('Inherited')
_SPECIALIZED_ROOT_PATH = Sdf.Path.absoluteRootPath.AppendChild('Specialized')
_BROADCAST_METHOD_RELPATHS = {Usd.Inherits: _INHERITED_ROOT_PATH, Usd.Specializes: _SPECIALIZED_ROOT_PATH}

_UNIT_UNIQUE_ID = ids.CGAsset.item  # Entry point for meaningful composed assets.
_UNIT_ORIGIN_PATH = Sdf.Path.absoluteRootPath.AppendChild("Origin")

# Composition filters for asset edits
_ASSET_UNIT_QUERY_FILTER = Usd.PrimCompositionQuery.Filter()
_ASSET_UNIT_QUERY_FILTER.dependencyTypeFilter = Usd.PrimCompositionQuery.DependencyTypeFilter.Direct
_ASSET_UNIT_QUERY_FILTER.hasSpecsFilter = Usd.PrimCompositionQuery.HasSpecsFilter.HasSpecs


def _fetch_layer(identifier: str, context: Ar.ResolverContext) -> Sdf.Layer:
    """Retrieve `layer <https://graphics.pixar.com/usd/docs/api/class_sdf_layer.html>`_ for the given ``identifier``.

     If the `layer <https://graphics.pixar.com/usd/docs/api/class_sdf_layer.html>`_ does not exist, it is created in the repository.
     """
    if not (layer := Sdf.Layer.Find(identifier) or Sdf.Layer.FindOrOpen(identifier)):
        # TODO: see how to make this repo_path better, seems very experimental atm.
        if context.IsEmpty():
            raise ValueError(f"Empty {context=} while fetching {identifier=}")

        repo_path = Path(context.Get()[0].GetSearchPath()[0])  # or just Repository.get()?
        Sdf.Layer.CreateNew(str(repo_path / identifier))
        if not (layer := Sdf.Layer.FindOrOpen(identifier)):
            raise RuntimeError(f"Make sure a resolver context with statement is being used. {context=}, {identifier=}")
    return layer


def _asset_identifier(path):
    # Expect identifiers to not have folders in between.
    if not path:
        raise ValueError("Can not extract asset identifier from empty path.")
    path = Path(path)
    if not path.is_absolute():
        return str(path)
    else:
        return str(path.relative_to(Repository.get()))


@typing.overload
def fetch_stage(identifier: str, context: Ar.ResolverContext = None, load=Usd.Stage.LoadAll) -> Usd.Stage:
    ...


@typing.overload
def fetch_stage(identifier: UsdAsset, context: Ar.ResolverContext = None, load=Usd.Stage.LoadAll) -> Usd.Stage:
    ...  # TODO: evaluate if it's worth to keep this, or if identifier can be a relative path


@functools.singledispatch
def fetch_stage(identifier, context: Ar.ResolverContext = None, load=Usd.Stage.LoadAll) -> Usd.Stage:
    """Retrieve the `stage <https://graphics.pixar.com/usd/docs/api/class_usd_stage.html>`_ whose root `layer <https://graphics.pixar.com/usd/docs/api/class_sdf_layer.html>`_ matches the given ``identifier``.

    If the `layer <https://graphics.pixar.com/usd/docs/api/class_sdf_layer.html>`_ does not exist, it is created in the repository.

    If an open matching `stage <https://graphics.pixar.com/usd/docs/api/class_usd_stage.html>`_ is found on the `global cache <https://graphics.pixar.com/usd/docs/api/class_usd_utils_stage_cache.html>`_, return it.
    Otherwise open it, populate the `cache <https://graphics.pixar.com/usd/docs/api/class_usd_utils_stage_cache.html>`_ and return it.

    .. attention::
        ``identifier`` must be a valid :class:`grill.names.UsdAsset` name.

    """
    if not context:
        context = Ar.ResolverContext(Ar.DefaultResolverContext([str(Repository.get())]))

    with Ar.ResolverContextBinder(context):
        layer = _fetch_layer(identifier, context)
        return Usd.Stage.Open(layer, load=load)


@fetch_stage.register(UsdAsset)
def _(identifier: UsdAsset, *args, **kwargs) -> Usd.Stage:
    return fetch_stage.registry[object](identifier.name, *args, **kwargs)


@fetch_stage.register(str)
def _(identifier: str, *args, **kwargs) -> Usd.Stage:
    return fetch_stage(UsdAsset(identifier), *args, **kwargs)


def define_taxon(stage: Usd.Stage, name: str, *, references: tuple[Usd.Prim] = tuple(), id_fields: typing.Mapping[str, str] = types.MappingProxyType({})) -> Usd.Prim:
    """Define a new `taxon group <https://en.wikipedia.org/wiki/Taxon>`_ for asset `taxonomy <https://en.wikipedia.org/wiki/Taxonomy>`_.

    If an existing ``taxon`` with the provided name already exists in the `stage <https://graphics.pixar.com/usd/docs/api/class_usd_stage.html>`_, it is used.

    The new ``taxon`` can extend from existing ``taxa`` via the ``references`` argument.

    Optional ``field=value`` items can be provided for identification purposes via ``id_fields``.

    :returns: `Prim <https://graphics.pixar.com/usd/docs/api/class_usd_prim.html>`_ representing the ``taxon`` group.
    """
    if name == _TAXONOMY_NAME:
        # TODO: prevent upper case lower case mismatch handle between multiple OS?
        #  (e.g. Windows considers both the same but Linux does not)
        raise ValueError(f"Can not define a taxon with reserved name: '{_TAXONOMY_NAME}'.")

    reserved_fields = {_TAXONOMY_UNIQUE_ID, _UNIT_UNIQUE_ID}
    reserved_fields.update([i.name for i in reserved_fields])
    if intersection:=reserved_fields.intersection(id_fields):
        raise ValueError(f"Can not provide reserved id fields: {', '.join(map(str, intersection))}. Got fields: {', '.join(map(str, id_fields))}")

    fields = {
        (token.name if isinstance(token, ids.CGAsset) else token): value
        for token, value in id_fields.items()
    }
    if invalid_fields:=set(fields).difference(ids.CGAsset.__members__):
        raise ValueError(f"Got invalid id_field keys: {', '.join(invalid_fields)}. Allowed: {', '.join(ids.CGAsset.__members__)}")

    with taxonomy_context(stage):
        prim = stage.DefinePrim(_TAXONOMY_ROOT_PATH.AppendChild(name))
        with Sdf.ChangeBlock():
            prim_references = prim.GetReferences()
            for reference in references:
                prim_references.AddInternalReference(reference.GetPath())
            prim.GetInherits().AddInherit(prim.GetPath().ReplacePrefix(Sdf.Path.absoluteRootPath, _INHERITED_ROOT_PATH))  # TODO: needed?
            taxon_fields = {**fields, _TAXONOMY_UNIQUE_ID.name: name}
            prim.SetAssetInfoByKey(_ASSETINFO_KEY, {_FIELDS_KEY: taxon_fields, _TAXA_KEY: {name: 0}})

    return prim


def itaxa(prims, taxon, *taxa):
    """Yields prims that are part of the given taxa."""
    taxa_names = {i if isinstance(i, str) else i.GetName() for i in (taxon, *taxa)}
    return (prim for prim in prims if taxa_names.intersection(prim.GetAssetInfoByKey(_ASSETINFO_TAXA_KEY) or {}))


def _catalogue_path(taxon: Usd.Prim) -> Sdf.Path:
    taxon_fields = _get_id_fields(taxon)
    # Check if we need to enforce _CATALOGUE_ID to be on taxon fields
    relpath = taxon_fields[_TAXONOMY_UNIQUE_ID.name]
    if _CATALOGUE_ID.name in taxon_fields:  # TODO: ensure this can't be overwritten
        relpath = f"{taxon_fields[_CATALOGUE_ID.name]}{Sdf.Path.childDelimiter}{relpath}"
    return _CATALOGUE_ROOT_PATH.AppendPath(relpath)


def _broadcast_root_path(taxon, broadcast_method, scope_path=None):
    scope_path = scope_path or _catalogue_path(taxon)  # TODO: this feels strange, avoid the "or" later.
    return scope_path.ReplacePrefix(_CATALOGUE_ROOT_PATH, _BROADCAST_METHOD_RELPATHS[broadcast_method])


def create_many(taxon, names, labels=tuple()) -> typing.List[Usd.Prim]:
    """Create a new taxon member for each of the provided names.

    When creating hundreds or thousands of members, this provides a considerable performance improvement over :func:`create_unit`.

    The new members will be created as `prims <https://graphics.pixar.com/usd/docs/api/class_usd_prim.html>`_ on the given ``taxon``'s `stage <https://graphics.pixar.com/usd/docs/api/class_usd_stage.html>`_.

    .. seealso:: :func:`define_taxon` :func:`create_unit`
    """
    stage = taxon.GetStage()
    taxon_path = taxon.GetPath()
    taxon_fields = _get_id_fields(taxon)
    scope_path = _catalogue_path(taxon)
    specialized_path = _broadcast_root_path(taxon, Usd.Specializes, scope_path=scope_path)
    inherited_path = _broadcast_root_path(taxon, Usd.Inherits, scope_path=scope_path)

    current_asset_name, root_layer = _root_asset(stage)
    new_asset_name = UsdAsset(current_asset_name.get(**taxon_fields))
    # Edits will go to the first layer that matches a valid pipeline identifier
    # TODO: Evaluate if this agreement is robust enough for different workflows.

    # existing = {i.GetName() for i in _iter_taxa(taxon.GetStage(), *taxon.GetCustomDataByKey(_ASSETINFO_TAXA_KEY))}
    taxonomy_layer = _find_layer_matching(_TAXONOMY_FIELDS, stage.GetLayerStack())
    taxonomy_id = _asset_identifier(taxonomy_layer.identifier)
    context = stage.GetPathResolverContext()
    if context.IsEmpty():  # Use a resolver context that is populated with the repository only when the context is empty.
        context = Ar.ResolverContext(Ar.DefaultResolverContext([str(Repository.get())]))

    try:
        catalogue_layer = _find_layer_matching(_CATALOGUE_FIELDS, stage.GetLayerStack())
    except ValueError:  # first time adding the catalogue layer
        catalogue_asset = current_asset_name.get(**_CATALOGUE_FIELDS)
        with Ar.ResolverContextBinder(context):
            catalogue_layer = _fetch_layer(str(catalogue_asset), context)
        catalogue_id = _asset_identifier(catalogue_layer.identifier)
        root_layer.subLayerPaths.insert(0, catalogue_id)
        # TODO: try setting this on session layer?

    # Some workflows like houdini might load layers without permissions to edit
    # Since we know we are in a valid pipeline layer, temporarily allow edits
    # for our operations, then restore original permissions.
    current_permission = catalogue_layer.permissionToEdit
    catalogue_layer.SetPermissionToEdit(True)

    scope = stage.GetPrimAtPath(scope_path)

    def _fetch_layer_for_unit(name):
        layer_id = str(new_asset_name.get(**{_UNIT_UNIQUE_ID.name: name}))
        layer = _fetch_layer(layer_id, context)
        origin = Sdf.CreatePrimInLayer(layer, _UNIT_ORIGIN_PATH)
        origin.specifier = Sdf.SpecifierDef
        origin.specializesList.Prepend(specialized_path.AppendChild(name))
        origin.inheritPathList.Prepend(inherited_path.AppendChild(name))
        origin.referenceList.Prepend(Sdf.Reference(taxonomy_id, taxon_path))
        layer.defaultPrim = origin.name
        return layer

    labels = itertools.chain(labels, itertools.repeat(""))
    with Usd.EditContext(stage, catalogue_layer), Ar.ResolverContextBinder(context):
        # Scope collecting all units based on taxon
        if not scope:
            scope = stage.DefinePrim(scope_path)
        if not scope.IsModel():
            # We use groups to ensure our scope is part of a valid model hierarchy
            for path in scope.GetPath().GetPrefixes():
                Usd.ModelAPI(stage.GetPrimAtPath(path)).SetKind(Kind.Tokens.group)

        # Place "homogeneus" operations under SdfChangeBlock to let composition notifications be sent only when needed.
        with Sdf.ChangeBlock():
            prims_info = []
            for name, label in zip(names, labels):
                if not stage.GetPrimAtPath(path:=scope_path.AppendChild(name)):
                    stage.OverridePrim(path)
                layer = _fetch_layer_for_unit(name)
                layer_id = _asset_identifier(layer.identifier)
                prims_info.append((name, label or name, path, layer, Sdf.Reference(layer_id)))

        prims_info = {stage.GetPrimAtPath(info[2]): info for info in prims_info}
        with Sdf.ChangeBlock():
            for prim, (name, label, *__, layer, reference) in prims_info.items():
                prim.GetReferences().AddReference(reference)
                with _usd.edit_context(reference, prim):
                    if hasattr(prim, "SetDisplayName"):  # USD-23.02+
                        prim.SetDisplayName(label)
                    UsdGeom.ModelAPI.Apply(prim)
                    modelAPI = Usd.ModelAPI(prim)
                    modelAPI.SetKind(Kind.Tokens.component)
                    modelAPI.SetAssetName(name)
                    modelAPI.SetAssetIdentifier(_asset_identifier(layer.identifier))

    catalogue_layer.SetPermissionToEdit(current_permission)
    return list(prims_info)


def create_unit(taxon: Usd.Prim, name: str, label: str = "") -> Usd.Prim:
    """Create a unit member of the given ``taxon``, with an optional display label.

    The new member will be created as a `prim <https://graphics.pixar.com/usd/docs/api/class_usd_prim.html>`_ on the given ``taxon``'s `stage <https://graphics.pixar.com/usd/docs/api/class_usd_stage.html>`_.

    .. seealso:: :func:`define_taxon` and :func:`create_many`
    """
    return create_many(taxon, [name], [label])[0]


def taxonomy_context(stage: Usd.Stage) -> Usd.EditContext:
    """Get an `edit context <https://graphics.pixar.com/usd/docs/api/class_usd_edit_context.html>`_ where edits will target this `stage <https://graphics.pixar.com/usd/docs/api/class_usd_stage.html>`_'s taxonomy `layer <https://graphics.pixar.com/usd/docs/api/class_sdf_layer.html>`_.

    .. attention::
        If a valid taxonomy `layer <https://graphics.pixar.com/usd/docs/api/class_sdf_layer.html>`_ is not found on the `layer stack <https://graphics.pixar.com/usd/docs/USD-Glossary.html#USDGlossary-LayerStack>`_, one is added to the `stage <https://graphics.pixar.com/usd/docs/api/class_usd_stage.html>`_.
    """
    # https://en.wikipedia.org/wiki/Taxonomy_(biology)
    try:
        taxonomy_layer = _find_layer_matching(_TAXONOMY_FIELDS, stage.GetLayerStack())
    except ValueError:
        # Our layer is not yet on the current layer stack. Let's bring it.
        # TODO: first valid pipeline layer is ok? or should it be current edit target?
        # TODO: try setting this on session layer?
        root_asset, root_layer = _root_asset(stage)
        taxonomy_asset = root_asset.get(**_TAXONOMY_FIELDS)
        context = stage.GetPathResolverContext()
        if context.IsEmpty():  # Use a resolver context that is populated with the repository only when the context is empty.
            context = Ar.ResolverContext(Ar.DefaultResolverContext([str(Repository.get())]))
        with Ar.ResolverContextBinder(context):
            taxonomy_layer = _fetch_layer(str(taxonomy_asset), context)
        # Use paths relative to our repository to guarantee portability
        # taxonomy_id = str(Path(taxonomy_layer.realPath).relative_to(Repository.get()))
        taxonomy_id = _asset_identifier(taxonomy_layer.identifier)
        taxonomy_root = Sdf.CreatePrimInLayer(taxonomy_layer, _TAXONOMY_ROOT_PATH)
        taxonomy_root.specifier = Sdf.SpecifierClass
        taxonomy_layer.defaultPrim = taxonomy_root.name
        root_layer.subLayerPaths.append(taxonomy_id)

    return Usd.EditContext(stage, taxonomy_layer)


def unit_context(prim: Usd.Prim) -> Usd.EditContext:
    """Get an `edit context <https://graphics.pixar.com/usd/docs/api/class_usd_edit_context.html>`_ where edits will target this `prim <https://graphics.pixar.com/usd/docs/api/class_usd_prim.html>`_'s unit root `layer <https://graphics.pixar.com/usd/docs/api/class_sdf_layer.html>`_."""
    # This targets the origin prim spec as the "entry point" edit target for a unit of a taxon.
    # This means some operations like specializes, inherits or internal reference / payloads
    # might not be able to be resolved (you will se an error like:
    # 'Cannot map </Catalogue/OtherPlace/CastleDracula> to current edit target.'
    layer = unit_asset(prim)

    def target_predicate(arc: Usd.CompositionArc):
        node = arc.GetTargetNode()
        return node.path == _UNIT_ORIGIN_PATH and node.layerStack.identifier.rootLayer == layer

    return _usd.edit_context(prim, _ASSET_UNIT_QUERY_FILTER, target_predicate)


def unit_asset(prim: Usd.Prim) -> Sdf.Layer:
    """Get the asset layer that acts as the 'entry point' for the given prim."""
    with Ar.ResolverContextBinder(prim.GetStage().GetPathResolverContext()):
        # Use Layer.Find since layer should have been open for the prim to exist.
        if layer:=Sdf.Layer.Find(Usd.ModelAPI(prim).GetAssetIdentifier().path):
            return layer
    fields = {**_get_id_fields(prim), _UNIT_UNIQUE_ID: Usd.ModelAPI(prim).GetAssetName()}
    return _find_layer_matching(fields, (i.layer for i in prim.GetPrimStack()))


def spawn_unit(parent, child, path=Sdf.Path.emptyPath, label=""):
    """Spawn a unit prim as a descendant of another.

    * Both parent and child must be existing units in the catalogue.
    * If `path <https://graphics.pixar.com/usd/docs/USD-Glossary.html#USDGlossary-Path>`_ is not provided, the name of child will be used.
    * A valid `Model Hierarchy <https://graphics.pixar.com/usd/docs/USD-Glossary.html#USDGlossary-ModelHierarchy>`_ is preserved by:

      1. Turning parent into an `assembly <https://graphics.pixar.com/usd/docs/USD-Glossary.html#USDGlossary-Assembly>`_.
      2. Ensuring intermediate prims between parent and child are also `models <https://graphics.pixar.com/usd/docs/USD-Glossary.html#USDGlossary-Model>`_.
      3. Setting explicit `instanceable <https://graphics.pixar.com/usd/docs/USD-Glossary.html#USDGlossary-Instanceable>`_. on spawned children that are components.

    .. seealso:: :func:`spawn_many` and :func:`create_unit`
    """
    return spawn_many(parent, child, [path or child.GetName()], [label])[0]


def spawn_many(parent: Usd.Prim, child: Usd.Prim, paths: list[Sdf.Path], labels: list[str] = []):
    """Spawn many instances of a prim unit as descendants of another.

    * Both parent and child must be existing units in the catalogue.
    * ``paths`` can be relative or absolute. If absolute, they must include ``parent``'s `path <https://graphics.pixar.com/usd/docs/USD-Glossary.html#USDGlossary-Path>`_ as a prefix.
    * A valid `Model Hierarchy <https://graphics.pixar.com/usd/docs/USD-Glossary.html#USDGlossary-ModelHierarchy>`_ is preserved by:

      1. Turning parent into an `assembly <https://graphics.pixar.com/usd/docs/USD-Glossary.html#USDGlossary-Assembly>`_ if ``child`` is a Model.
      2. Ensuring intermediate prims between ``parent`` and spawned children are also `models <https://graphics.pixar.com/usd/docs/USD-Glossary.html#USDGlossary-Model>`_.
      3. Setting explicit `instanceable <https://graphics.pixar.com/usd/docs/USD-Glossary.html#USDGlossary-Instanceable>`_. on spawned children that are components.

    .. seealso:: :func:`spawn_unit` and :func:`create_unit`
    """
    if parent == child:
        raise ValueError(f"Can not spawn {parent} on to itself.")
    parent_path = parent.GetPath()
    paths_to_create = []
    for path in paths:
        if isinstance(path, str):
            path = Sdf.Path(path)
        if path.IsAbsolutePath():  # If path is an absolute path, fail if it sits outside of parent's path.
            if not path.HasPrefix(parent_path) or path == parent_path:
                raise ValueError(f"{path=} needs to be a child path of parent path {parent_path}")
        else:
            path = parent_path.AppendPath(path)
        paths_to_create.append(path)
    labels = itertools.chain(labels, itertools.repeat(""))
    try:
        reference = _asset_identifier(Usd.ModelAPI(child).GetAssetIdentifier().path)
    except ValueError:
        raise ValueError(f"Could not extract identifier from {child} to spawn under {parent}.")
    parent_stage = parent.GetStage()
    spawned = [parent_stage.DefinePrim(path) for path in paths_to_create]
    child_is_model = child.IsModel()
    with Sdf.ChangeBlock():
        # Action of bringing a unit from our catalogue turns parent into an assembly only if child is a model.
        if child_is_model and not (parent_model := Usd.ModelAPI(parent)).IsKind(Kind.Tokens.assembly):
            parent_model.SetKind(Kind.Tokens.assembly)
        for spawned_unit, label in zip(spawned, labels):
            # Use reference for the asset to:
            # 1. Make use of instancing as much as possible with fewer prototypes.
            # 2. Let specializes / inherits changes later.
            spawned_unit.GetReferences().AddReference(reference)
            if label and hasattr(spawned_unit, "SetDisplayName"):  # USD-23.02+
                spawned_unit.SetDisplayName(label)
            # Action of bringing a unit from our catalogue turns parent into an assembly only if child is a model.
            if child_is_model:
                # check for all intermediate parents of our spawned unit to ensure valid model hierarchy
                for inner_parent in _usd.iprims(parent_stage, [parent_path], lambda p: p == spawned_unit.GetParent()):
                    if not inner_parent.IsModel():
                        Usd.ModelAPI(inner_parent).SetKind(Kind.Tokens.group)
                if not child.IsGroup():
                    # Sensible defaults: component prims are instanced
                    spawned_unit.SetInstanceable(True)
    return spawned


def _root_asset(stage):
    """From a give stage, find the first layer that matches a valid grill identifier.

    This can be useful in situations when a stage's root layer is not a valid identifier (e.g. anonymous) but
    has sublayered a valid one in the pipeline.

    This searches on the root layer first.
    """
    with contextlib.suppress(ValueError):
        root_layer = stage.GetRootLayer()
        return UsdAsset(Path(root_layer.identifier).name), root_layer

    seen = set()
    for layer in stage.GetLayerStack():
        try:
            return UsdAsset(Path(layer.identifier).name), layer
        except ValueError:
            seen.add(layer)
    raise ValueError(f"Could not find a valid pipeline layer for stage {stage}. Searched layer stack: {pformat(seen)}")


def _get_id_fields(prim):
    if not (fields:=prim.GetAssetInfoByKey(_ASSETINFO_FIELDS_KEY)):
        raise ValueError(f"Missing or empty '{_FIELDS_KEY}' on '{_ASSETINFO_KEY}' asset info for {prim}. Got: {pformat(prim.GetAssetInfoByKey(_ASSETINFO_KEY))}")
    if not isinstance(fields, typing.Mapping):
        raise TypeError(f"Expected mapping on key '{_FIELDS_KEY}' from {prim} on custom data key '{_ASSETINFO_KEY}'. Got instead {fields} with type: {type(fields)}")
    return fields


def _find_layer_matching(tokens: typing.Mapping, layers: typing.Iterable[Sdf.Layer]) -> Sdf.Layer:
    """Find the first layer matching the given identifier tokens.

    :raises ValueError: If none of the given layers match the provided tokens.
    """
    tokens = {
        ((token.name if isinstance(token, ids.CGAsset) else token), value)
        for token, value in tokens.items()
    }
    seen = set()
    for layer in layers:
        # anonymous layers realPath defaults to an empty string
        name = UsdAsset(Path(layer.realPath).name)
        if tokens.difference(name.values.items()):
            seen.add(layer)
            continue
        return layer
    raise ValueError(f"Could not find layer matching {tokens}. Searched on:\n{pformat(seen)}")


def specialized_context(prim, context_unit=None):
    return _inherit_or_specialize_unit(prim.GetSpecializes(), context_unit)


def inherited_context(prim, context_unit=None):
    return _inherit_or_specialize_unit(prim.GetInherits(), context_unit)


def _inherit_or_specialize_unit(method, context_unit):
    """This is on cook since it relies on some pipeline knowledge to find proper target. Could request the target
    path as well at the expense of the caller if need arises or enough value is perceived."""
    target_prim = method.GetPrim()
    if not (unit_name:=Usd.ModelAPI(target_prim).GetAssetName()):
        raise ValueError(f"{target_prim} is not a valid unit in the catalogue.")

    context_unit = context_unit or target_prim
    if not Usd.ModelAPI(context_unit).GetAssetName():
        raise ValueError(f"{context_unit=} needs to be a valid unit in the catalogue.")

    broadcast_method = type(method)
    if not target_prim.GetPath().HasPrefix(context_unit.GetPath()):
        raise ValueError(f"Can not check for {broadcast_method} on {context_unit} since {target_prim} is not a descendant of it.")

    target_path = _broadcast_root_path(target_prim, broadcast_method).AppendPath(unit_name)

    try:
        return _usd.edit_context(method, target_path, (context_asset:=unit_asset(context_unit)))
    except ValueError as exc:
        raise ValueError(
            f"Could not find an appropriate edit target node for a {broadcast_method.__name__}'s arc targeting {target_path} for {target_prim}. "
            f"""Is there a composition arc bringing "{target_prim.GetName()}"'s unit into "{context_unit.GetName()}"'s layer stack at {context_asset}?"""
        ) from exc


def taxonomy_graph(prims, url_id_prefix):
    graph = nx.DiGraph(tooltip="Taxonomy Graph")
    graph.graph.update(
        graph={'rankdir': 'LR'},
        node={
            'shape': 'box',
            'fillcolor': "#afd7ff",  # lightskyblue1
            'color': "#1E90FF",  # dodgerblue4
            'style': 'filled,rounded',
        },
    )

    # TODO:
    #  - Guarantee taxa will be unique (no duplicated short names), raise here?
    #  - Fail with clear error message when provided prims are not taxa
    for taxon in prims:
        graph.add_node(taxon_name:=taxon.GetName(), tooltip=taxon.GetPath(), href=f"{url_id_prefix}{taxon_name}",)
        graph.add_edges_from(itertools.zip_longest(set(taxon.GetAssetInfoByKey(_ASSETINFO_TAXA_KEY)) - {taxon_name}, (), fillvalue=taxon_name))
    return graph



# USDView not on pypi yet, so not possible to test this on CI
import types
import inspect
import operator
import contextvars
from functools import lru_cache, partial

from pxr import UsdGeom, Usd, Sdf, Ar, Tf
from pxr.Usdviewq import plugin, layerStackContextMenu, attributeViewContextMenu, primContextMenuItems, primContextMenu

import grill.usd as gusd

from ._qt import QtWidgets, QtGui
from . import _core, _attributes, sheets as _sheets, description as _description, create as _create, stats as _stats

_usdview_api = contextvars.ContextVar("_usdview_api")  # TODO: is there a better way?
_description._PALETTE.set(0)  # TODO 2: same question (0 == dark, 1 == light)


def _findOrCreateMenu(parent, title):
    return next((child for child in parent.findChildren(QtWidgets.QMenu) if child.title() == title), None) or parent.addMenu(title)


def _addAction(self, *args, **kwargs):
    if len(args) == 2:
        # primContextMenu.PrimContextMenu calls addAction(menuItem.GetText(), menuItem.RunCommand)
        # This will break as soon as it's called differently, but it's a risk worth to take for now.
        path, method = args
        if inspect.ismethod(method) and isinstance(method.__self__, (_GrillPrimContextMenuItem, GrillAttributeEditorMenuItem)):
            path_segments = path.split("|")
            if len(path_segments) > 2:
                raise RuntimeError(f"Don't know how to handle submenus larger than 2: {path_segments}")
            if len(path_segments) == 2:
                submenu, action_text = path_segments
                child_menu = _findOrCreateMenu(self, submenu)
                if action_text == "...":
                    for text, runner in method.__self__._GetSubCommands():
                        child_menu.addAction(text, runner)
                    return
                return child_menu.addAction(action_text, method)

    return super(type(self), self).addAction(*args, **kwargs)


primContextMenu.PrimContextMenu.addAction = _addAction
attributeViewContextMenu.AttributeViewContextMenu.addAction = _addAction


def _stage_on_widget(widget_creator):
    @lru_cache(maxsize=None)
    def _launcher(usdviewApi):
        widget = widget_creator(parent=usdviewApi.qMainWindow)
        widget.setStage(usdviewApi.stage)
        return widget
    return _launcher


def _layer_stack_from_prims(usdviewApi):
    widget = _description.LayerStackComposition(parent=usdviewApi.qMainWindow)
    widget.setPrimPaths(usdviewApi.dataModel.selection.getPrimPaths())
    widget.setStage(usdviewApi.stage)
    return widget


@lru_cache(maxsize=None)
def prim_composition(usdviewApi):
    widget = _description.PrimComposition(parent=usdviewApi.qMainWindow)

    def primChanged(new_paths, __):
        widget.setPrim(next(map(usdviewApi.stage.GetPrimAtPath, new_paths), None))

    usdviewApi.dataModel.selection.signalPrimSelectionChanged.connect(primChanged)
    if usdviewApi.prim:
        widget.setPrim(usdviewApi.prim)
    return widget


def _connection_viewer(usdviewApi):
    widget = _description._ConnectableAPIViewer(parent=usdviewApi.qMainWindow)

    def primChanged(new_paths, __):
        widget.setPrim(next(map(usdviewApi.stage.GetPrimAtPath, new_paths), None))

    usdviewApi.dataModel.selection.signalPrimSelectionChanged.connect(primChanged)
    if usdviewApi.prim:
        widget.setPrim(usdviewApi.prim)
    return widget


def save_changes(usdviewApi):
    def show():
        if QtWidgets.QMessageBox.question(
            usdviewApi.qMainWindow, "Save Changes", "All changes will be saved to disk.\n\nContiue?"
        ) == QtWidgets.QMessageBox.Yes:
            usdviewApi.stage.Save()
    return types.SimpleNamespace(show=show)


def repository_path(usdviewApi):
    show = partial(_create.CreateAssets._setRepositoryPath, usdviewApi.qMainWindow)
    return types.SimpleNamespace(show=show)


class GrillContentBrowserLayerMenuItem(layerStackContextMenu.LayerStackContextMenuItem):
    def IsEnabled(self):
        # Layer Stack Tab provides `layerPath`. Composition provides `layer`. Try both.
        return bool(getattr(self._item, 'layer', None) or getattr(self._item, 'layerPath', None))

    def GetText(self):
        return _description._BROWSE_CONTENTS_MENU_TITLE

    def RunCommand(self):
        if self._item:
            # from testing, only entries in the Composition tab without HasSpecs miss "path"
            if path := getattr(self._item, "path", None):
                if isinstance(path, str):
                    path = Sdf.Path(path)
                paths = [path]
            else:
                paths = []
            usdview_api = _usdview_api.get()
            context = usdview_api.stage.GetPathResolverContext()
            if not (layer:= getattr(self._item, 'layer', None)):  # USDView allows for single layer selection in composition tab :(
                layerPath = getattr(self._item, 'layerPath', "")
                # We're protected by the IsEnabled method above, so don't bother checking layerPath value
                with Ar.ResolverContextBinder(context):
                    if not (layer:=Sdf.Layer.FindOrOpen(layerPath)):  # edge case, is this possible?
                        print(f"Could not find layer from {layerPath}")
                        return
            _description._launch_content_browser([layer], usdview_api.qMainWindow, context, paths=paths)


class _GrillPrimContextMenuItem(primContextMenuItems.PrimContextMenuItem):
    """A prim context menu item class that allows special Grill behavior like being added to submenus."""
    _items = []

    def __init_subclass__(cls, **kwargs):
        _GrillPrimContextMenuItem._items.append(cls)


class GrillPrimCompositionMenuItem(_GrillPrimContextMenuItem):
    _widget = _description.PrimComposition
    _subtitle = "Prim Composition"

    def GetText(self):
        return f"Inspect|{self._subtitle}"

    def RunCommand(self):
        usdview_api = _usdview_api.get()
        # The "double pop up" upon showing widgets does not happen on PySide2, only on PySide6
        for prim in self._selectionDataModel.getPrims():
            widget = self._widget(parent=usdview_api.qMainWindow)
            widget.setPrim(prim)
            widget.show()


class GrillLayerStackCompositionMenuItem(GrillPrimCompositionMenuItem):
    _widget = _description.LayerStackComposition
    _subtitle = "LayerStack Composition"

    def RunCommand(self):
        usdview_api = _usdview_api.get()
        stage = usdview_api.stage
        prims = self._selectionDataModel.getPrims()
        widget = self._widget(parent=usdview_api.qMainWindow)
        widget.setPrimPaths(prim.GetPath() for prim in prims)
        widget.setStage(stage)
        widget.show()


class GrillPrimConnectionViewerMenuItem(GrillPrimCompositionMenuItem):
    _widget = _description._ConnectableAPIViewer
    _subtitle = "Connections"


class AllHierarchyTextMenuItem(_GrillPrimContextMenuItem):
    _include_descendants = True
    _subtitle = "All Descendants"
    _predicate = Usd.TraverseInstanceProxies(Usd.PrimAllPrimsPredicate)

    def GetText(self):
        return f"Copy Hierarchy|{self._subtitle}"

    def RunCommand(self):
        text = gusd._format_prim_hierarchy(self._selectionDataModel.getPrims(), self._include_descendants, self._predicate)
        clipboard = QtWidgets.QApplication.clipboard()
        clipboard.setText(text, QtGui.QClipboard.Selection)
        clipboard.setText(text, QtGui.QClipboard.Clipboard)


class ModelHierarchyTextMenuItem(AllHierarchyTextMenuItem):
    _subtitle = "Model Hierarchy"
    _predicate = Usd.PrimIsModel


class SelectedHierarchyTextMenuItem(AllHierarchyTextMenuItem):
    _include_descendants = False
    _subtitle = "Selection Only"


class GrillAttributeEditorMenuItem(attributeViewContextMenu.AttributeViewContextMenuItem):
    @property
    def _attributes(self):
        return [i for i in self._dataModel.selection.getProps() if isinstance(i, Usd.Attribute)]

    def ShouldDisplay(self):
        return (
            self._role == attributeViewContextMenu.PropertyViewDataRoles.ATTRIBUTE and
            attr.GetMetadata('allowedTokens') if len(self._attributes) == 1 and (attr := self._attributes[0]).GetTypeName() == Sdf.ValueTypeNames.Token else True
        )

    def IsEnabled(self):
        return self._item and self._attributes

    def GetText(self):
        if (selected := len(self._attributes)) == 1 and self._attributes[0].GetTypeName() in {Sdf.ValueTypeNames.Bool, Sdf.ValueTypeNames.Token}:
            return "Set Value|..."
        return f"Edit Value{'s' if selected > 1 else ''}"

    def RunCommand(self):
        if attributes:=self._attributes:
            editor = _ValueEditor(_usdview_api.get().qMainWindow)
            editor.setAttributes(attributes)
            editor.show()

    def _GetSubCommands(self):
        """Collect value options to provide as menu actions when an attribute of a supported types is selected."""
        attribute, = self._attributes
        type_name = attribute.GetTypeName()
        if type_name == Sdf.ValueTypeNames.Bool:
            return [(str(value), partial(attribute.Set, value)) for value in (True, False)]
        elif type_name == Sdf.ValueTypeNames.Token:
            tokens = attribute.GetMetadata('allowedTokens')
            return [(value, partial(attribute.Set, value)) for value in tokens]

class _ValueEditor(QtWidgets.QDialog):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        layout = QtWidgets.QFormLayout(self)
        self.setLayout(layout)
        self.setWindowTitle("Experimental Value Editor")

    def setAttributes(self, attributes):
        layout = self.layout()
        supported_primvars = {"displayColor"}
        for attr in attributes:
            print(attr)
            type_name = attr.GetTypeName()
            if (primvar:= UsdGeom.Primvar(attr)) and primvar.GetPrimvarName() in supported_primvars:
                editor = _attributes._DisplayColorEditor(primvar)
                layout.addRow(primvar.GetPrimvarName(), editor)
            elif type_name == Sdf.ValueTypeNames.Token:
                tokens = attr.GetMetadata('allowedTokens') or []  # unregistered tokens could return a None value
                def update(what, value):
                    what.Set(value)
                editor = QtWidgets.QComboBox(self)
                editor.addItems(tokens)
                if (current := attr.Get()) in set(tokens):
                    editor.setCurrentText(current)
                elif not tokens:
                    msg = "No 'allowedTokens' registered"
                    editor.addItems([msg])
                    editor.setCurrentText(msg)
                    editor.setEnabled(False)
                layout.addRow(attr.GetName(), editor)
                editor.currentTextChanged.connect(partial(update, attr))
            elif type_name == Sdf.ValueTypeNames.String:
                def update(what, editor):
                    what.Set(editor.text())
                editor = QtWidgets.QLineEdit(self)
                editor.setText(attr.Get() or "")
                layout.addRow(attr.GetName(), editor)
                editor.returnPressed.connect(partial(update, attr, editor))
            elif type_name in {Sdf.ValueTypeNames.Double, Sdf.ValueTypeNames.Float}:
                def update(what, value):
                    what.Set(value)
                editor = QtWidgets.QDoubleSpinBox(self)
                editor.setValue(attr.Get())
                layout.addRow(attr.GetName(), editor)
                editor.valueChanged.connect(partial(update, attr))
            elif type_name == Sdf.ValueTypeNames.Bool:
                editor = QtWidgets.QCheckBox(self)
                editor.setChecked(attr.Get())
                layout.addRow(attr.GetName(), editor)
                def update(ed, what, *__):
                    what.Set(ed.isChecked())
                editor.stateChanged.connect(partial(update, editor, attr))
            else:
                print(f"Don't know how to edit {attr} of type {type_name}")


class GrillPlugin(plugin.PluginContainer):

    def registerPlugins(self, plugRegistry, usdviewApi):
        _usdview_api.set(usdviewApi)

        def show(_launcher, _usdviewAPI):
            return _launcher(_usdviewAPI).show()

        def _menu_item(title, _launcher):
            # contract: _launcher() returns an object that shows a widget on `show()`
            return plugRegistry.registerCommandPlugin(
                f"Grill.{title.replace(' ', '_')}", title, partial(show, _launcher),
            )

        self._menu_items = [
            *(_menu_item(title, launcher) for (title, launcher) in (
                ("Create Assets", _stage_on_widget(_create.CreateAssets)),
                ("Taxonomy Editor", _stage_on_widget(_create.TaxonomyEditor)),
                ("Spreadsheet Editor", _stage_on_widget(_sheets.SpreadsheetEditor)),
                ("Prim Composition", prim_composition),
                ("Connection Viewer", _connection_viewer),
            )),
            {"LayerStack Composition": [
                _menu_item("From Current Stage", _stage_on_widget(_description.LayerStackComposition)),
                _menu_item("From Selected Prims", _layer_stack_from_prims),
            ]},
            operator.methodcaller("addSeparator"),
            *(_menu_item(title, launcher) for title, launcher in (
                ("Stage Stats", _stage_on_widget(_stats.StageStats)),
                ("Save Changes", save_changes),
            )),
            operator.methodcaller("addSeparator"),
            {"Preferences": [_menu_item("Repository Path", repository_path)],},
        ]

    def configureView(self, plugRegistry, plugUIBuilder):
        def _populate_menu(menu, items):
            for item in items:
                if isinstance(item, operator.methodcaller):
                    item(menu)
                elif isinstance(item, dict):
                    for child_menu_name, child_items in item.items():
                        child_menu = menu.findOrCreateSubmenu(child_menu_name)
                        _populate_menu(child_menu, child_items)
                else:
                    menu.addItem(item)
        grill_menu = plugUIBuilder.findOrCreateMenu("👨‍🍳 Grill")
        _populate_menu(grill_menu, self._menu_items)


def _extend_menu(_extender, original, *args):
    return [extension(*args) for extension in _extender] + original(*args)  # if it looks like a duck


for module, member_name, extender in (
        (primContextMenuItems, "_GetContextMenuItems", _GrillPrimContextMenuItem._items),
        (layerStackContextMenu, "_GetContextMenuItems", (GrillContentBrowserLayerMenuItem,)),
        # _GetContextMenuItems(item, dataModel) signature is inverse than GrillAttributeEditorMenuItem(dataModel, item)
        (attributeViewContextMenu, "_GetContextMenuItems", (lambda *args: GrillAttributeEditorMenuItem(*reversed(args)),))
):
    setattr(module, member_name, partial(_extend_menu, extender, getattr(module, member_name)))


# We need to do this since primContextMenu imports the function directly, so re-assign with our recently patched one
primContextMenu._GetContextMenuItems = primContextMenuItems._GetContextMenuItems
Tf.Type.Define(GrillPlugin)


def _patch_style(cls):  # while a nicer way comes up
    original = cls.__init__

    def _init_with_usdview_style(self, *args, **kwargs):
        original(self, *args, **kwargs)
        self.setStyleSheet(_core._USDVIEW_STYLE)
    cls.__init__ = _init_with_usdview_style


for cls in _description._Tree, _description.LayerStackComposition, _sheets._StageSpreadsheet:
    # Only when in USDView we want to extend the stylesheet of some of the classes tha require them
    # TODO: find a better way to do this
    _patch_style(cls)



import os
import json
import logging
from pathlib import Path

from ..resources import houdini

logger = logging.getLogger(__name__)


def install_package():
    """Install The Grill package on the users houdini preferences.

    In the future this might be automatic so no extra user calls should be required.
    """
    logger = logging.getLogger(__name__)
    user_pref_dir = os.getenv('HOUDINI_USER_PREF_DIR')
    if not os.getenv('HOUDINI_USER_PREF_DIR'):
        raise RuntimeError("Can execute only when a houdini environment with HOUDINI_USER_PREF_DIR is available. Aborting.")
    logger.debug(f"Installing houdini preferences to {user_pref_dir}")
    # 1. Get the source json package for the grill
    resources_path = Path(houdini.__path__._path[0])
    pkg_src = resources_path / 'package.json'
    if not pkg_src.is_file():
        raise RuntimeError(f"Aborting without changes. Could not locate expected package file at {pkg_src}. Try re-installing from PyPi.")
    logger.debug(f"pkg_src = {pkg_src}")
    # 2. Update the source data with the new directory location
    pkg_data = json.loads(pkg_src.read_bytes())
    logger.debug(f"pkg_obj = {pkg_data}")
    pkg_data['env'][0]['HOUDINI_MENU_PATH']['value'] = str(resources_path)
    logger.debug(f"pkg_obj = {pkg_data}")
    # 3. Save updated data in user preferences
    pkg_tgt = Path(user_pref_dir) / "packages" / "grill.json"
    pkg_tgt.parent.mkdir(parents=True, exist_ok=True)
    pkg_tgt.write_text(json.dumps(pkg_data, indent=4))
    print(f"Successfully installed package on {pkg_tgt}")



"""Views related to USD scene description"""
from __future__ import annotations

import re
import typing
import weakref
import logging
import operator
import tempfile
import contextvars
import collections

from pathlib import Path
from itertools import chain
from collections import defaultdict
from functools import cache, partial
from types import MappingProxyType

import networkx as nx
from pxr import UsdUtils, UsdShade, Usd, Ar, Pcp, Sdf, Tf
from ._qt import QtWidgets, QtGui, QtCore

from .. import usd as _usd
from . import sheets as _sheets, _core, _graph
from ._core import _which


_logger = logging.getLogger(__name__)

_color_attrs = lambda color: dict.fromkeys(("color", "fontcolor"), color)
_ARCS_LEGEND = MappingProxyType({
    Pcp.ArcTypeInherit: _color_attrs('mediumseagreen'),
    Pcp.ArcTypeVariant: _color_attrs('orange'),
    Pcp.ArcTypeRelocate: _color_attrs('indigo'),  # ~indigo
    Pcp.ArcTypeReference: _color_attrs('crimson'),  # ~red
    Pcp.ArcTypePayload: _color_attrs('#9370db'),  # ~purple
    Pcp.ArcTypeSpecialize: _color_attrs('sienna'),  # ~brown
})
_BROWSE_CONTENTS_MENU_TITLE = 'Browse Contents'
_PALETTE = contextvars.ContextVar("_PALETTE", default=1)  # (0 == dark, 1 == light)
_HIGHLIGHT_COLORS = MappingProxyType(
    {name: (QtGui.QColor(dark), QtGui.QColor(light)) for name, (dark, light) in (
        ("comment", ("gray", "gray")),
        # A mix between:
        # https://en.wikipedia.org/wiki/Solarized#Colors
        # https://www.w3schools.com/colors/colors_palettes.asp
        *((key, ("#f7786b", "#d33682")) for key in ("specifier", "rel_op", "value_assignment", "references", "reference")),
        *((key, ("#f7cac9", "#f7786b")) for key in ("prim_name", "identifier_prim_path", "relationship", "apiSchemas")),
        *((key, ("#b5e7a0", "#859900")) for key in ("metadata", "interpolation_meta", "custom_meta")),
        *((key, ("#ffcc5c", "#b58900")) for key in ("identifier", "variantSets", "variantSet", "variants", "variant", "prop_name", "rel_name", "outline_details")),
        *((key, (_ARCS_LEGEND[Pcp.ArcTypeInherit]['color'], _ARCS_LEGEND[Pcp.ArcTypeInherit]['color'])) for key in ("inherit", "inherits")),
        *((key, ("#9a94bf", _ARCS_LEGEND[Pcp.ArcTypePayload]['color'])) for key in ("payload", "payloads")),
        *((key,  ("#c18f76", _ARCS_LEGEND[Pcp.ArcTypeSpecialize]['color'])) for key in ("specialize", "specializes")),
        ("list_op", ("#e3eaa7", "#2aa198")),
        *((key, ("#5b9aa0", "#5b9aa0")) for key in ("prim_type", "prop_type", "prop_array")),
        *((key, ("#87bdd8", "#034f84")) for key in ("set_string", "string_value")),
        ("collapsed", ("#92a8d1", "#36486b")),
        ("boolean", ("#b7d7e8", "#50394c")),
        ("number", ("#bccad6", "#667292")),
    )}
)
_HIGHLIGHT_PATTERN = re.compile(
    rf'(^(?P<comment>#.*$)|^( *(?P<specifier>def|over|class)( (?P<prim_type>\w+))? (?P<prim_name>\"\w+\")| +((?P<metadata>(?P<arc_selection>variants|payload|references)|{"|".join(_usd._metadata_keys())})|(?P<list_op>add|(ap|pre)pend|delete) (?P<arc>inherits|variantSets|references|payload|specializes|apiSchemas|rel (?P<rel_name>[\w:]+))|(?P<variantSet>variantSet) (?P<set_string>\"\w+\")|(?P<custom_meta>custom )?(?P<interpolation_meta>uniform )?(?P<prop_type>{"|".join(_usd._attr_value_type_names())}|dictionary|rel)(?P<prop_array>\[])? (?P<prop_name>[\w:.]+))( (\(|((?P<value_assignment>= )[\[(]?))|$))|(?P<string_value>\"[^\"]+\")|(?P<identifier>@[^@]+@)(?P<identifier_prim_path><[/\w]+>)?|(?P<relationship><[/\w:.]+>)|(?P<collapsed><< [^>]+ >>)|(?P<boolean>true|false)|(?P<number>-?[\d.]+))'
)

_OUTLINE_SDF_PATTERN = re.compile(  # this is a very minimal draft to have colors on the outline sdffilter mode.
    rf'^((?P<identifier_prim_path>(  )?</[/\w+.:{{}}=]*(\[[/\w+.:{{}}=]+])?>)( : (?P<specifier>\w+))?|(  )?(?P<metadata>\w+)(: ((?P<number>-?[\d.]+)|(?P<string_value>[\w:.]+)(?P<prop_array>\[])?|(?P<collapsed><< [^>]+ >>)|(\[ (?P<relationship>[\w /.:{{}}=]+) ])|(?P<outline_details>.+)))?)$'  # sad, last item is a dot, lazy atm
)

_TREE_PATTERN = re.compile(  # draft as well output of usdtree
    rf'^(?P<identifier_prim_path>([/`|\s:]+-+)(\w+)?)(\((?P<metadata>\w+)\)|(?P<prop_name>\.[\w:]+)|( \[(?P<specifier>def|over|class)( (?P<prim_type>\w+))?])( \((?P<custom_meta>[\w\s=]+)\))?)'
)

_USD_COMPOSITION_ARC_QUERY_METHODS = (
    Usd.CompositionArc.HasSpecs,
    Usd.CompositionArc.IsAncestral,
    Usd.CompositionArc.IsImplicit,
    Usd.CompositionArc.IsIntroducedInRootLayerPrimSpec,
    Usd.CompositionArc.IsIntroducedInRootLayerStack,
)
_USD_COMPOSITION_ARC_QUERY_KEYS = tuple(func.__name__ for func in _USD_COMPOSITION_ARC_QUERY_METHODS)
_USD_COMPOSITION_ARC_QUERY_DEFAULTS = lambda: dict.fromkeys(_USD_COMPOSITION_ARC_QUERY_KEYS, False)


@cache
def _edge_color(edge_arcs):
    return dict(color=":".join(_ARCS_LEGEND[arc]["color"] for arc in edge_arcs))


def _format_layer_contents(layer, output_type="pseudoLayer", paths=tuple(), output_args=tuple()):
    """Textual representation of a layer using ``sdffilter``."""
    with tempfile.TemporaryDirectory() as target_dir:
        name = Path(layer.realPath).stem if layer.realPath else "".join(c if c.isalnum() else "_" for c in layer.identifier)
        path = Path(target_dir) / f"{name}.usd"
        try:
            layer.Export(str(path))
        except Tf.ErrorException:
            # Prefer crate export for performance, although it could fail to export non-standard layers.
            # When that fails, try export with original file format.
            # E.g. content that fails to export: https://github.com/PixarAnimationStudios/OpenUSD/blob/59992d2178afcebd89273759f2bddfe730e59aa8/pxr/usd/sdf/testenv/testSdfParsing.testenv/baseline/127_varyingRelationship.sdf#L9
            path = path.with_suffix(f".{layer.fileExtension}")
            layer.Export(str(path))
        path_args = ("-p", "|".join(re.escape(str(p)) for p in paths)) if paths else tuple()
        if output_type == "usdtree":
            args = [_which("usdtree"), "-a", "-m", str(path)]
        else:
            args = [_which("sdffilter"), "--outputType", output_type, *output_args, *path_args, str(path)]
        return _core._run(args)


def _layer_label(layer):
    return layer.GetDisplayName() or layer.identifier


@cache
def _highlight_syntax_format(key, value):
    text_fmt = QtGui.QTextCharFormat()
    if key == "arc":
        key = "rel_op" if value.startswith("rel") else value
    elif key == "arc_selection":
        key = value
    elif key == "comment":
        text_fmt.setFontItalic(True)
    elif key == "specifier":
        if value == "over":
            text_fmt.setFontItalic(True)
        elif value == "class":
            text_fmt.setFontLetterSpacing(135)
    elif key == "identifier":
        text_fmt.setFontUnderline(True)
    text_fmt.setForeground(_HIGHLIGHT_COLORS[key][_PALETTE.get()])
    return text_fmt


def _compute_layerstack_graph(prims, url_prefix) -> _GraphInfo:
    """Compute layer stack graph info for the provided prims"""

    @cache
    def _sublayers(layer_stack):  # {Sdf.Layer: int}
        return MappingProxyType({v: i for i, v in enumerate(layer_stack.layers)})

    def _add_node(pcp_node):
        layer_stack = pcp_node.layerStack
        root_layer = layer_stack.identifier.rootLayer
        sublayers = _sublayers(layer_stack)
        try:  # lru_cache not compatible with Pcp objects, so we "cache" at the layer level
            return ids_by_root_layer[root_layer], sublayers
        except KeyError:
            pass  # layerStack still not processed, let's add it
        ids_by_root_layer[root_layer] = index = len(all_nodes)

        attrs = dict(style='rounded,filled', shape='record', href=f"{url_prefix}{index}", fillcolor="white", color="darkslategray")
        plugs = []
        label = '{'
        tooltip = 'LayerStack:'
        for layer, layer_index in sublayers.items():
            indices_by_sublayers[layer].add(index)
            if layer.dirty:
                attrs['color'] = 'darkorange'  # can probably be dashed as well?
            # For new line: https://stackoverflow.com/questions/16671966/multiline-tooltip-for-pydot-graph
            # For Windows path sep: https://stackoverflow.com/questions/15094591/how-to-escape-forwardslash-character-in-html-but-have-it-passed-correctly-to-jav
            tooltip += f"&#10;{layer_index}: {(layer.realPath or layer.identifier)}".replace('\\', '&#47;')
            plugs.append(layer_index)
            label += f"{'' if layer_index == 0 else '|'}<{layer_index}>{_layer_label(layer)}"
        label += '}'
        # attrs['plugs'] = dict(zip(plugs, range(len(plugs))))
        attrs['plugs'] = tuple(plugs)
        attrs['active_plugs'] = set()  # all active connections, for GUI
        all_nodes[index] = dict(label=label, tooltip=tooltip, **attrs)
        return index, sublayers

    @cache
    def _compute_composition(_prim):
        query = Usd.PrimCompositionQuery(_prim)
        affected_by = set()  # {int}  indices of nodes affecting this prim
        for arc in query.GetCompositionArcs():
            target_idx, __ = _add_node(arc.GetTargetNode())
            affected_by.add(target_idx)
            source_layer = arc.GetIntroducingLayer()
            if source_layer:
                # Note: arc.GetIntroducingNode() is not guaranteed to be the same as
                # arc.GetTargetNode().origin nor arc.GetTargetNode().GetOriginRootNode()
                source_idx, source_layers = _add_node(arc.GetIntroducingNode())
                source_port = source_layers[source_layer]
                all_nodes[source_idx]['active_plugs'].add(source_port)  # all connections, for GUI
                all_edges[source_idx, target_idx][source_port][arc.GetArcType()].update(
                    {func.__name__:  is_fun for func in _USD_COMPOSITION_ARC_QUERY_METHODS if (is_fun := func(arc))}
                )

        return affected_by

    all_nodes = dict()  # {int: dict}
    all_edges = defaultdict(                            #   { (source_node: int, target_node: int):
        lambda: defaultdict(                            #       { source_port: int:
            lambda: defaultdict(                        #           { Pcp.ArcType:
                _USD_COMPOSITION_ARC_QUERY_DEFAULTS     #               { HasArcs: bool, IsImplicit: bool, ... }
            )                                           #           }
        )                                               #       }
    )                                                   #   }

    for arc_type, attributes in _ARCS_LEGEND.items():
        arc_label_node_ids = (len(all_nodes), len(all_nodes) + 1)
        all_nodes.update(dict.fromkeys(arc_label_node_ids, dict(style='invis')))
        all_edges[arc_label_node_ids][None][arc_type] = dict(label=f" {arc_type.displayName}", **attributes)

    legend_node_ids = tuple(all_nodes)
    ids_by_root_layer = dict()
    indices_by_sublayers = defaultdict(set)  # {Sdf.Layer: {int,} }
    paths_by_node_idx = defaultdict(set)

    for prim in prims:
        # use a forwarded prim in case of instanceability to avoid computing same stack more than once
        prim_path = prim.GetPath()
        for affected_by_idx in _compute_composition(UsdUtils.GetPrimAtPathWithForwarding(prim.GetStage(), prim_path)):
            paths_by_node_idx[affected_by_idx].add(prim_path)

    return _GraphInfo(
        edges=MappingProxyType(all_edges),
        nodes=MappingProxyType(all_nodes),
        sticky_nodes=legend_node_ids,
        paths_by_ids=paths_by_node_idx,
        ids_by_layers=indices_by_sublayers,
    )


def _graph_from_connections(prim: Usd.Prim) -> nx.MultiDiGraph:
    connections_api = UsdShade.ConnectableAPI(prim)
    graph = nx.MultiDiGraph()
    outline_color = "#4682B4"  # 'steelblue'
    background_color = "#F0FFFF"  # 'azure'
    graph.graph['graph'] = {'rankdir': 'LR'}

    graph.graph['node'] = {'shape': 'none', 'color': outline_color, 'fillcolor': background_color}  # color and fillcolor used for HTML view
    graph.graph['edge'] = {"color": 'crimson'}

    all_nodes = dict()  # {node_id: {graphviz_attr: value}}
    edges = list()  # [(source_node_id, target_node_id, {source_plug_name, target_plug_name, graphviz_attrs})]

    @cache
    def _get_node_id(api):
        return str(api.GetPrim().GetPath())

    @cache
    def _add_edges(src_node, src_name, tgt_node, tgt_name):
        tooltip = f"{src_node}.{src_name} -> {tgt_node}.{tgt_name}"
        edges.append((src_node, tgt_node, {"tailport": src_name, "headport": tgt_name, "tooltip": tooltip}))

    plug_colors = {
        UsdShade.Input: outline_color,  # blue
        UsdShade.Output: "#F08080"  # "lightcoral",  # pink
    }
    table_row = '<tr><td port="{port}" border="0" bgcolor="{color}" style="ROUNDED">{text}</td></tr>'

    traversed_prims = set()
    def traverse(api: UsdShade.ConnectableAPI):
        current_prim = api.GetPrim()
        if current_prim in traversed_prims:
            return
        traversed_prims.add(current_prim)
        node_id = _get_node_id(current_prim)
        label = f'<<table border="1" cellspacing="2" style="ROUNDED" bgcolor="{background_color}" color="{outline_color}">'
        label += table_row.format(port="", color="white", text=f'<font color="{outline_color}"><b>{api.GetPrim().GetName()}</b></font>')
        plugs = {"": 0}  # {graphviz port name: port index order}
        active_plugs = set()
        for index, plug in enumerate(chain(api.GetInputs(), api.GetOutputs()), start=1):  # we start at 1 because index 0 is the node itself
            plug_name = plug.GetBaseName()
            sources, __ = plug.GetConnectedSources()  # (valid, invalid): we care only about valid sources (index 0)
            color = plug_colors[type(plug)] if isinstance(plug, UsdShade.Output) or sources else background_color
            label += table_row.format(port=plug_name, color=color, text=f'<font color="#242828">{plug_name}</font>')
            for source in sources:
                _add_edges(_get_node_id(source.source.GetPrim()), source.sourceName, node_id, plug_name)
                traverse(source.source)
            plugs[plug_name] = index
            active_plugs.add(plug_name)  # TODO: add only actual plugged properties, right now we're adding all of them
        label += '</table>>'
        all_nodes[node_id] = dict(label=label, plugs=plugs, active_plugs=active_plugs)

    traverse(connections_api)

    graph.add_nodes_from(all_nodes.items())
    graph.add_edges_from(edges)
    return graph


def _launch_content_browser(layers, parent, context, paths=tuple()):
    dialog = _start_content_browser(layers, parent, context, paths)
    dialog.show()


def _start_content_browser(layers, parent, context, paths=tuple()):
    dialog = QtWidgets.QDialog(parent=parent)
    dialog.setWindowTitle("Layer Content Browser")
    layout = QtWidgets.QVBoxLayout()
    vertical = QtWidgets.QSplitter(QtCore.Qt.Vertical)
    for layer in layers:
        browser = _PseudoUSDBrowser(layer, parent=dialog, resolver_context=context, paths=paths)
        vertical.addWidget(browser)
    layout.addWidget(vertical)
    dialog.setLayout(layout)
    return dialog


class _GraphInfo(typing.NamedTuple):
    sticky_nodes: tuple
    ids_by_layers: typing.Mapping
    edges: typing.Mapping
    nodes: typing.Mapping
    paths_by_ids: typing.Mapping


@cache
def _nx_graph_edge_filter(*, has_specs=None, ancestral=None, implicit=None, introduced_in_root_layer_prim_spec=None, introduced_in_root_layer_stack=None):
    match = {
        key: value for key, value in (
            ("HasSpecs", has_specs),
            ("IsAncestral", ancestral),
            ("IsImplicit", implicit),
            ("IsIntroducedInRootLayerPrimSpec", introduced_in_root_layer_prim_spec),
            ("IsIntroducedInRootLayerStack", introduced_in_root_layer_stack),
        ) if value is not None
    }
    if not match:
        return None
    return lambda edge_info: match.items() <= edge_info.items()


class _ConnectableAPIViewer(QtWidgets.QDialog):
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent=parent, **kwargs)
        self._graph_view = _graph._GraphViewer(parent=self)
        vertical = QtWidgets.QSplitter(QtCore.Qt.Vertical)
        vertical.addWidget(self._graph_view)
        self.setFocusProxy(self._graph_view)
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(vertical)
        self.setLayout(layout)

    def setPrim(self, prim):
        if not prim:
            return
        prim = prim.GetPrim()
        type_text = "" if not (type_name := prim.GetTypeName()) else f" {type_name}"
        self.setWindowTitle(f"Scene Graph Connections From{type_text}: {prim.GetName()} ({prim.GetPath()})")
        self._graph_view.graph = graph = _graph_from_connections(prim)
        if isinstance(self._graph_view, _graph._GraphSVGViewer):
            self._graph_view.view(graph.nodes.keys())


# Reminder: Inheriting does not bring QTreeView stylesheet (Stylesheet needs to target this class specifically).
class _Tree(_core._ColumnHeaderMixin, QtWidgets.QTreeView):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.expanded.connect(self._expand_all_children)
        self.collapsed.connect(self._collapse_all_children)

    def _expand_all_children(self, index):
        if QtWidgets.QApplication.keyboardModifiers() == QtCore.Qt.ShiftModifier:
            with QtCore.QSignalBlocker(self):
                self.expandRecursively(index)

    def _collapse_all_children(self, index):
        model = index.model()

        def _collapse_recursively(_index):  # No "self.collapseRecursively", Qt? ):
            self.setExpanded(_index, False)
            for row in range(model.rowCount(_index)):
                if child := model.index(row, 0, _index):
                    _collapse_recursively(child)

        if QtWidgets.QApplication.keyboardModifiers() == QtCore.Qt.ShiftModifier:
            with QtCore.QSignalBlocker(self):
                _collapse_recursively(index)

    def _connect_search(self, options, index, model):
        super()._connect_search(options, index, model)
        model.setRecursiveFilteringEnabled(True)
        options.filterChanged.connect(self.expandAll)


class PrimComposition(QtWidgets.QDialog):
    # TODO: See if columns need to be updated from dict to tuple[_core.Column]
    _COLUMNS = {
        "Target Layer": lambda arc: _layer_label(arc.GetTargetNode().layerStack.layerTree.layer),
        "Arc": lambda arc: arc.GetArcType().displayName,
        "#": lambda arc: arc.GetTargetNode().siblingNumAtOrigin,
        "Target Path": lambda arc: arc.GetTargetNode().path,
        "Has Specs": Usd.CompositionArc.HasSpecs,
        "Is Ancestral": Usd.CompositionArc.IsAncestral,
        "Is Implicit": Usd.CompositionArc.IsImplicit,
        "From Root Layer Prim Spec": Usd.CompositionArc.IsIntroducedInRootLayerPrimSpec,
        "From Root LayerStack": Usd.CompositionArc.IsIntroducedInRootLayerStack,
    }

    def __init__(self, *args, **kwargs):
        """For inspection and debug purposes, this widget makes primary use of:

            - Usd.PrimCompositionQuery  (similar to USDView's composition tab)
            - Pcp.PrimIndex.DumpToString
            - Pcp.PrimIndex.DumpToDotGraph  (when dot is available)
        """
        super().__init__(*args, **kwargs)
        self.index_box = QtWidgets.QTextBrowser()
        self.index_box.setLineWrapMode(QtWidgets.QTextBrowser.NoWrap)
        self._composition_model = model = QtGui.QStandardItemModel()
        columns = tuple(_core._Column(k, v) for k, v in self._COLUMNS.items())
        options = _core._ColumnOptions.SEARCH
        self.composition_tree = tree = _Tree(model, columns, options)
        tree.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        tree.customContextMenuRequested.connect(self._exec_context_menu)
        tree.setAlternatingRowColors(True)
        tree.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)

        self._dot_view = _graph._DotViewer(parent=self)

        tree_controls = QtWidgets.QFrame()
        tree_controls_layout = QtWidgets.QHBoxLayout()
        tree_controls.setLayout(tree_controls_layout)
        self._prim = None  # TODO: see if we can remove this. Atm needed for "enabling layer stack" checkbox
        self._complete_target_layerstack = QtWidgets.QCheckBox("Complete Target LayerStack")
        self._complete_target_layerstack.setChecked(False)
        self._complete_target_layerstack.clicked.connect(lambda: self.setPrim(self._prim))
        tree_controls_layout.addWidget(self._complete_target_layerstack)
        tree_controls_layout.addStretch(0)
        tree_controls_layout.setContentsMargins(0,0,0,0)
        tree_controls.setContentsMargins(0,0,0,0)
        tree_controls.setFixedHeight(tree_controls.sizeHint().height())

        vertical = QtWidgets.QSplitter(QtCore.Qt.Vertical)
        vertical.addWidget(tree_controls)
        vertical.addWidget(tree)
        vertical.addWidget(self._dot_view)
        vertical.setStretchFactor(2, 1)
        horizontal = QtWidgets.QSplitter(QtCore.Qt.Horizontal)
        horizontal.addWidget(vertical)
        horizontal.addWidget(self.index_box)
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(horizontal)
        self.setLayout(layout)
        self.setWindowTitle("Prim Composition")

    def clear(self):
        self._composition_model.clear()
        self.index_box.clear()

    def _exec_context_menu(self):
        # https://doc.qt.io/qtforpython-5/overviews/statemachine-api.html#the-state-machine-framework
        menu = QtWidgets.QMenu(tree:=self.composition_tree)
        selection = tree.selectedIndexes()
        if len(set(i.row() for i in selection) ) == 1:
            stage, edit_target, target_path = selection[0].data(QtCore.Qt.UserRole)
            menu.addAction("Set As Edit Target", partial(stage.SetEditTarget, edit_target))
            menu.addAction(_BROWSE_CONTENTS_MENU_TITLE, partial(_launch_content_browser, (edit_target.GetLayer(),), self, stage.GetPathResolverContext(), [target_path]))
        menu.exec_(QtGui.QCursor.pos())

    def setPrim(self, prim):
        self._prim = prim
        if not self._prim:
            self.clear()
            self.setWindowTitle("Prim Composition")
            return
        self.setWindowTitle(f"Prim Composition: {prim.GetName()} ({prim.GetPath()})")
        prim_index = prim.GetPrimIndex()
        self.index_box.setText(prim_index.DumpToString())
        fd, fp = tempfile.mkstemp()
        prim_index.DumpToDotGraph(fp)
        self._dot_view.setDotPath(fp)

        complete_target_layerstack = self._complete_target_layerstack.isChecked()
        # model.clear() resizes the columns. So we keep track of current sizes.
        header = self.composition_tree.header()
        sizes = {index: size for index in header.options_by_index if (size:=header.sectionSize(index))}

        tree = self.composition_tree
        model = self._composition_model
        model.clear()
        model.setHorizontalHeaderLabels([""] * len(self._COLUMNS))
        root_item = model.invisibleRootItem()

        query = Usd.PrimCompositionQuery(prim)
        items = dict()
        stage = prim.GetStage()
        arcs = query.GetCompositionArcs()
        for arc in arcs:
            values = [getter(arc) for getter in self._COLUMNS.values()]

            intro_node = arc.GetIntroducingNode()
            target_node = arc.GetTargetNode()
            target_path = target_node.path
            target_layer = target_node.layerStack.identifier.rootLayer

            # TODO: is there a better way than relying on str(node.site)???
            key = str(intro_node.site)
            parent = items[key] if key in items else root_item

            try:
                highlight_color = _HIGHLIGHT_COLORS[arc.GetArcType().displayName][_PALETTE.get()]
            except KeyError:
                highlight_color = None

            sublayers = target_node.layerStack.layers if complete_target_layerstack else (target_layer,)
            for each in sublayers:
                if each == target_layer:  # we're the root layer of the target node's stack
                    arc_items = [QtGui.QStandardItem(str(s)) for s in values]
                    items[str(target_node.site)] = arc_items[0]
                else:
                    has_specs = bool(each.GetObjectAtPath(target_path))
                    arc_items = [QtGui.QStandardItem(str(s)) for s in [_layer_label(each), values[1], values[2], values[3], str(has_specs)]]

                edit_target = Usd.EditTarget(each, target_node)
                for item in arc_items:
                    item.setData((stage, edit_target, target_path), QtCore.Qt.UserRole)
                    if highlight_color:
                        item.setData(highlight_color, QtCore.Qt.ForegroundRole)

                parent.appendRow(arc_items)

        tree.expandAll()
        tree._fixPositions()  # TODO: Houdini needs this. Why?
        for index, size in sizes.items():
            header.resizeSection(index, size)


class LayerTableModel(_core._ObjectTableModel):
    def data(self, index:QtCore.QModelIndex, role:int=...) -> typing.Any:
        if role == QtCore.Qt.ForegroundRole:
            layer = self.data(index, role=_core._QT_OBJECT_DATA_ROLE)
            # TODO: add a concrete color value for "dirty" layers?
            color = _sheets._PrimTextColor.ARCS if layer.dirty else _sheets._PrimTextColor.NONE
            return color.value
        return super().data(index, role)

    def setLayers(self, value):
        self.beginResetModel()
        self._objects = list(value)
        self.endResetModel()


class _Highlighter(QtGui.QSyntaxHighlighter):
    _pattern = _HIGHLIGHT_PATTERN

    def highlightBlock(self, text):
        for match in re.finditer(self._pattern, text):
            for syntax_group, value in match.groupdict().items():
                if not value:
                    continue
                start, end = match.span(syntax_group)
                self.setFormat(start, end-start, _highlight_syntax_format(syntax_group, value))


class _SdfOutlineHighlighter(_Highlighter):
    _pattern = _OUTLINE_SDF_PATTERN


class _TreeOutlineHighlighter(_Highlighter):
    _pattern = _TREE_PATTERN


class _LayersSheet(_sheets._Spreadsheet):
    def __init__(self, *args, resolver_context=Ar.GetResolver().CreateDefaultContext(), **kwargs):
        super().__init__(*args, **kwargs)
        self._resolver_context = resolver_context

    def contextMenuEvent(self, event):
        if not _which("sdffilter"):
            print(
                "In order to display layer contents, the 'sdffilter' command must be \n"
                "available on the current environment.\n\n"
                "Please ensure 'sdffilter' is on the system's PATH environment variable."
            )
            return  # ATM only appear if sdffilter is in the environment
        self.menu = QtWidgets.QMenu(self)
        self.menu.addAction(_BROWSE_CONTENTS_MENU_TITLE, self._display_contents)
        self.menu.popup(QtGui.QCursor.pos())

    def _display_contents(self, *_, **__):
        selected = self.table.selectedIndexes()
        layers = {index.data(_core._QT_OBJECT_DATA_ROLE) for index in selected}
        _launch_content_browser(layers, self, self._resolver_context)


class _PseudoUSDBrowser(QtWidgets.QTabWidget):
    def __init__(self, layer, *args, resolver_context=Ar.GetResolver().CreateDefaultContext(), paths=tuple(), **kwargs):
        super().__init__(*args, **kwargs)
        self._resolver_context = resolver_context
        self._browsers_by_layer = dict()  # {Sdf.Layer: _PseudoUSDTabBrowser}
        self._tab_layer_by_idx = list()  # [tab_idx: Sdf.Layer]
        self._addLayerTab(layer, paths)
        self._resolved_layers = {layer}
        self.setTabsClosable(True)
        self.tabCloseRequested.connect(lambda idx: self.removeTab(idx))

    def tabRemoved(self, index: int) -> None:
        item = self._tab_layer_by_idx[index]
        if isinstance(item, (Sdf.Layer, weakref.ref)):
            item = item.__repr__.__self__
            self._resolved_layers.discard(item)
        del self._browsers_by_layer[self._tab_layer_by_idx.pop(index)]

    def mousePressEvent(self, event):
        if event.button() == QtCore.Qt.RightButton and (tab_index := self.tabBar().tabAt(event.pos())) != -1:
            self._menu_for_tab(tab_index).exec_(event.globalPos())

        super().mousePressEvent(event)

    def _menu_for_tab(self, tab_index):
        widget = self.widget(tab_index)
        clipboard = QtWidgets.QApplication.instance().clipboard()
        menu = QtWidgets.QMenu(self)
        menu.addAction("Copy Identifier", partial(clipboard.setText, widget._identifier))
        menu.addAction("Copy Resolved Path", partial(clipboard.setText, widget._resolved_path))
        menu.addSeparator()
        if tab_index < (max_tab_idx := len(self._tab_layer_by_idx)) - 1:
            menu.addAction("Close Tabs to the Right", partial(self._close_many, range(tab_index + 1, max_tab_idx + 1)))
        if tab_index > 0:
            menu.addAction("Close Tabs to the Left", partial(self._close_many, range(tab_index)))
        return menu

    def _close_many(self, indices: range):
        for index in reversed(indices):
            self.tabCloseRequested.emit(index)

    @_core.wait()
    def _addImageTab(self, path, *, identifier):
        try:
            focus_widget = self._browsers_by_layer[path]
        except KeyError:
            pixmap = QtGui.QPixmap(path)
            if pixmap.isNull():
                QtWidgets.QMessageBox.warning(self, "Error Opening Contents", f"Could not load {path}")
                return
            image_label = QtWidgets.QLabel(parent=self)
            image_label.setAlignment(QtCore.Qt.AlignCenter)
            image_label.resize(pixmap.size())
            image_label.setPixmap(pixmap)

            focus_widget = QtWidgets.QFrame(parent=self)
            focus_layout = QtWidgets.QHBoxLayout()
            focus_layout.setContentsMargins(0, 0, 0, 0)

            image_item = QtWidgets.QGraphicsPixmapItem(pixmap)
            viewport = _graph._GraphicsViewport(parent=self)
            scene = QtWidgets.QGraphicsScene()
            scene.addItem(image_item)
            viewport.setScene(scene)

            focus_layout.addWidget(viewport)
            focus_widget.setLayout(focus_layout)

            tab_idx = self.addTab(focus_widget, Path(path).name)

            focus_widget._resolved_path = path
            focus_widget._identifier = identifier
            self.setTabToolTip(tab_idx, path)

            self._tab_layer_by_idx.append(path)
            assert len(self._tab_layer_by_idx) == (tab_idx + 1)
            self._browsers_by_layer[path] = focus_widget

        self.setCurrentWidget(focus_widget)

    @_core.wait()
    def _addLayerTab(self, layer, paths=tuple(), *, identifier=None):
        layer_ref = weakref.ref(layer)
        try:
            focus_widget = self._browsers_by_layer[layer_ref]
        except KeyError:
            paths_in_layer = []
            for path in paths:
                if not layer.GetObjectAtPath(path):
                    _logger.debug(f"{path=} does not exist on {layer=}")
                    continue
                if path.IsPropertyPath():
                    path = path.GetParentPath()
                paths_in_layer.append(path)
            error, text = _format_layer_contents(layer, paths=paths_in_layer)
            if error:
                QtWidgets.QMessageBox.warning(self, "Error Opening Contents", error)
                return

            outliner_columns = (_core._Column("Path", lambda path: path.name),)
            outline_model = QtGui.QStandardItemModel()
            outline_tree = _Tree(outline_model, outliner_columns, _core._ColumnOptions.SEARCH)
            outline_tree.setSelectionMode(outline_tree.SelectionMode.ExtendedSelection)
            outline_tree.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)

            def show_outline_tree_context_menu(*args):
                if selected_indexes:= outline_tree.selectedIndexes():
                    content = "\n".join(str(index.data(QtCore.Qt.UserRole)) for index in selected_indexes if index.isValid())
                    menu = QtWidgets.QMenu(outline_tree)
                    menu.addAction("Copy Paths", partial(QtWidgets.QApplication.instance().clipboard().setText, content))
                    menu.exec_(QtGui.QCursor.pos())

            outline_tree.customContextMenuRequested.connect(show_outline_tree_context_menu)

            outline_model.setHorizontalHeaderLabels([""] * len(outliner_columns))
            root_item = outline_model.invisibleRootItem()

            items = dict()  # {Sdf.Path: QtGui.QItem}

            def populate(spec_paths, *_, **__):
                for path in spec_paths:
                    if path.IsPropertyPath() or path.IsTargetPath():
                        continue
                    parent_key = path.GetParentPath()
                    variant_set, selection = path.GetVariantSelection()
                    if path.IsPrimVariantSelectionPath() and selection:  # place all variant selections under the variant set
                        parent_key = parent_key.AppendVariantSelection(variant_set, "")
                    highlight_color = _HIGHLIGHT_COLORS["variantSets"][_PALETTE.get()] if variant_set else None
                    parent = items[parent_key] if parent_key in items else root_item
                    new_items = [QtGui.QStandardItem(str(column.getter(path))) for column in outliner_columns]
                    if highlight_color:
                        for item in new_items:
                            item.setData(highlight_color, QtCore.Qt.ForegroundRole)
                    parent.appendRow(new_items)
                    new_items[0].setData(path, QtCore.Qt.UserRole)
                    items[path] = new_items[0]

            content_paths = list()
            layer.Traverse(layer.pseudoRoot.path, lambda path: content_paths.append(path))
            selection_model = outline_tree.selectionModel()
            highligther_cls = _Highlighter
            highlighters = {"pseudoLayer": _Highlighter, "outline": _SdfOutlineHighlighter, "usdtree": _TreeOutlineHighlighter}
            def _ensure_highligther(cls):
                nonlocal highligther_cls
                if highligther_cls != cls:
                    browser.setText("")  # clear contents before changing highlighting to avoid locks with huge contents
                    highligther_cls = cls
                    highligther_cls(browser)
                sorting_combo.setEnabled(cls == _SdfOutlineHighlighter)
                outline_valies_check.setEnabled(cls == _SdfOutlineHighlighter)

            def update_contents(*_, **__):
                layer_ = layer_ref()
                format_choice = format_combo.currentText()
                output_args = []
                if format_choice == "pseudoLayer":
                    output_args += ["--arraySizeLimit", "6", "--timeSamplesSizeLimit", "6"]
                elif format_choice == "outline":
                    output_args += ["--sortBy", sorting_combo.currentText()]
                    if not outline_valies_check.isChecked():
                        output_args.append("--noValues")
                selected_indices = selection_model.selectedIndexes()
                paths = list()
                for each in selected_indices:
                    path = each.data(QtCore.Qt.UserRole)
                    variant_set, selection = path.GetVariantSelection()
                    if path.IsPrimVariantSelectionPath() and not selection:  # we're the "parent" variant. collect all variant paths as sdffilter does not math unselected variants ):
                        paths.extend([v.path for v in layer_.GetObjectAtPath(path).variants.values()])
                    else:
                        paths.append(path)

                with QtCore.QSignalBlocker(browser):
                    _ensure_highligther(highlighters.get(format_choice, _Highlighter))
                    error, text = _format_layer_contents(layer_, format_combo.currentText(), paths, output_args)
                    line_count = len(text.split("\n"))
                    line_counter.setText("\n".join(chain(map(str, range(1, line_count)), ["\n"] * 5)))
                    line_counter.setFixedWidth(12 + (len(str(line_count)) * 8))
                    browser.setText(error if error else text)

            populate(sorted(content_paths))  # Sdf.Layer.Traverse collects paths from deepest -> highest. Sort from high -> deep

            if paths_in_layer:
                tree_model = outline_tree.model()
                with QtCore.QSignalBlocker(outline_tree):
                    for path in paths_in_layer:
                        proxy_index = tree_model.mapFromSource(outline_model.indexFromItem(items[path]))
                        selection_model.select(proxy_index, QtCore.QItemSelectionModel.Select | QtCore.QItemSelectionModel.Rows)
                        outline_tree.setCurrentIndex(proxy_index)  # possible to set multiple?
            else:
                outline_tree.expandRecursively(root_item.index(), 3)

            focus_widget = QtWidgets.QFrame(parent=self)
            focus_splitter = QtWidgets.QSplitter(QtCore.Qt.Horizontal)
            focus_splitter.addWidget(outline_tree)
            focus_layout = QtWidgets.QHBoxLayout()
            focus_layout.setContentsMargins(0, 0, 0, 0)

            options_layout = QtWidgets.QHBoxLayout()
            format_layout = QtWidgets.QFormLayout()
            format_combo = QtWidgets.QComboBox()
            format_combo.addItems(["pseudoLayer", "outline", "usdtree"])
            focus_widget._format_options = format_combo
            format_layout.addRow("Format", format_combo)
            options_layout.addLayout(format_layout)

            outline_sorting_layout = QtWidgets.QFormLayout()
            sorting_combo = QtWidgets.QComboBox()
            sorting_combo.addItems(["path", "field"])
            outline_sorting_layout.addRow("Sort By", sorting_combo)
            options_layout.addLayout(outline_sorting_layout)
            outline_valies_check = QtWidgets.QCheckBox("Values")
            outline_valies_check.setChecked(True)
            options_layout.addWidget(outline_valies_check)
            options_layout.addStretch()

            format_combo.setCurrentIndex(0)
            sorting_combo.setEnabled(False)
            outline_valies_check.setEnabled(False)

            browser_frame = QtWidgets.QFrame(parent=self)
            browser_layout = QtWidgets.QVBoxLayout()
            browser_layout.setContentsMargins(0, 0, 0, 0)
            browser_layout.addLayout(options_layout)

            self._line_filter = browser_line_filter = QtWidgets.QLineEdit()
            browser_line_filter.setPlaceholderText("Find")

            filter_layout = QtWidgets.QFormLayout()
            filter_layout.addRow(_core._EMOJI.SEARCH.value, browser_line_filter)
            browser_layout.addLayout(filter_layout)

            browser_frame.setLayout(browser_layout)
            for trigger in (
                    outline_valies_check.toggled,
                    selection_model.selectionChanged,
                    format_combo.currentIndexChanged,
                    sorting_combo.currentIndexChanged,
            ):
                trigger.connect(update_contents)

            focus_layout.addWidget(focus_splitter)
            focus_splitter.addWidget(browser_frame)
            focus_splitter.setStretchFactor(1, 2)
            focus_widget.setLayout(focus_layout)

            browser = _PseudoUSDTabBrowser(parent=self)
            browser.setLineWrapMode(QtWidgets.QTextBrowser.NoWrap)
            _Highlighter(browser)
            browser.identifier_requested.connect(partial(self._on_identifier_requested, weakref.proxy(layer)))
            browser.setText(text)

            def _find(text):
                if text and not browser.find(text):
                    browser.moveCursor(QtGui.QTextCursor.Start)  # try again from start
                    browser.find(text)

            browser_line_filter.textChanged.connect(_find)
            browser_line_filter.returnPressed.connect(lambda: _find(browser_line_filter.text()))

            line_counter = QtWidgets.QTextBrowser()
            line_counter.setLineWrapMode(QtWidgets.QTextBrowser.NoWrap)
            line_counter.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
            line_counter.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

            line_counter_doc = line_counter.document()
            line_text_option = line_counter_doc.defaultTextOption()
            line_text_option.setAlignment(QtCore.Qt.AlignRight)
            line_counter_doc.setDefaultTextOption(line_text_option)

            line_count = len(text.split("\n"))
            line_counter.setFixedWidth(12 + (len(str(line_count)) * 8))
            line_counter.setText("\n".join(chain(map(str, range(1, line_count)), ["\n"]*5)))
            line_counter.setEnabled(False)

            line_scrollbar = line_counter.verticalScrollBar()
            browser.verticalScrollBar().valueChanged.connect(line_scrollbar.setValue)
            browser._line_counter = line_counter

            browser_combined_layout = QtWidgets.QHBoxLayout()
            browser_combined_layout.addWidget(line_counter)
            browser_combined_layout.setSpacing(0)
            browser_combined_layout.setContentsMargins(0,0,0,0)
            browser_combined_layout.addWidget(browser)

            browser_layout.addLayout(browser_combined_layout)

            tab_idx = self.addTab(focus_widget, _layer_label(layer))
            focus_widget._resolved_path = str(layer.resolvedPath)
            focus_widget._identifier = identifier or layer.identifier
            self.setTabToolTip(tab_idx, str(layer.resolvedPath))

            self._tab_layer_by_idx.append(layer_ref)
            assert len(self._tab_layer_by_idx) == (tab_idx+1)
            self._browsers_by_layer[layer_ref] = focus_widget
        self.setCurrentWidget(focus_widget)

    def _on_identifier_requested(self, anchor: Sdf.Layer, identifier: str):
        anchor = anchor.__repr__.__self__
        with Ar.ResolverContextBinder(self._resolver_context):
            try:
                if not (layer := Sdf.Layer.FindOrOpen(identifier)):
                    layer = Sdf.Layer.FindOrOpenRelativeToLayer(anchor, identifier)
            except Tf.ErrorException as exc:
                resolved_path = str(Ar.GetResolver().Resolve(identifier)) or anchor.ComputeAbsolutePath(identifier)
                if resolved_path and Path(resolved_path).suffix[1:] in _image_formats_to_browse():
                    self._addImageTab(resolved_path, identifier=identifier)
                    return
                title = "Error Opening File"
                text = str(exc.args[0])
            else:
                if layer:
                    self._addLayerTab(layer, identifier=identifier)
                    self._resolved_layers.add(layer)
                    return
                title = "Layer Not Found"
                text = f"Could not find layer with {identifier=} under resolver context {self._resolver_context} with {anchor=}"
            QtWidgets.QMessageBox.warning(self, title, text)

@cache
def _image_formats_to_browse():
    return frozenset(str(fmt, 'utf-8') for fmt in QtGui.QImageReader.supportedImageFormats())

class _PseudoUSDTabBrowser(QtWidgets.QTextBrowser):
    # See: https://doc.qt.io/qt-5/qtextbrowser.html#navigation
    # The anchorClicked() signal is emitted when the user clicks an anchor.
    # we should be able to use anchor functionality but that does not seem to work with syntax highlighting ):
    # https://stackoverflow.com/questions/35858340/clickable-hyperlink-in-qtextedit/61722734#61722734
    # https://www.qtcentre.org/threads/26332-QPlainTextEdit-and-anchors
    # https://stackoverflow.com/questions/66931106/make-all-matches-links-by-pattern
    # https://fossies.org/dox/CuteMarkEd-0.11.3/markdownhighlighter_8cpp_source.html
    identifier_requested = QtCore.Signal(str)
    _line_counter = None

    def mousePressEvent(self, event):
        cursor = self.cursorForPosition(event.pos())
        cursor.select(QtGui.QTextCursor.WordUnderCursor)
        word = cursor.selectedText()
        cursor.movePosition(QtGui.QTextCursor.StartOfLine, QtGui.QTextCursor.KeepAnchor)
        # Take advantage that identifiers in pseudo sdffilter come always in separate lines
        if f"{cursor.selectedText()}{word}".count('@') == 1 and (
                match := re.search(
                    rf"@(?P<identifier>[^@]*({re.escape(word.strip('@'))})[^@]*)@",
                    cursor.block().text()
                )
        ):
            self._target = match.group("identifier")
            QtWidgets.QApplication.setOverrideCursor(QtGui.Qt.PointingHandCursor)
        else:
            self._target = None
            super().mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        if self._target:
            QtWidgets.QApplication.restoreOverrideCursor()
            self.identifier_requested.emit(self._target)
            self._target = None
        else:
            super().mouseReleaseEvent(event)

    def wheelEvent(self, event):
        if event.modifiers() == QtCore.Qt.ControlModifier:
            method = operator.methodcaller("zoomIn" if event.angleDelta().y() > 0 else "zoomOut")
            method(self)
            if self._line_counter:
                method(self._line_counter)
        else:
            super().wheelEvent(event)


class LayerStackComposition(QtWidgets.QDialog):
    # TODO: display total amount of layer stacks, and sites
    #       TOO SLOW!! (ALab workbench environment takes 7 seconds for the complete stage.)
    _LAYERS_COLUMNS = (
        _core._Column(f"{_core._EMOJI.ID.value} Layer Identifier", operator.attrgetter('identifier')),
        _core._Column("🚧 Dirty", operator.attrgetter('dirty')),
    )
    _PRIM_COLUMNS = (
        _core._Column("🧩 Opinion on Prim Path", lambda prim: str(prim.GetPath())),
        _core._Column(f"{_core._EMOJI.NAME.value} Prim Name", Usd.Prim.GetName),
    )

    def __init__(self, parent=None, **kwargs):
        super().__init__(parent=parent, **kwargs)
        options = _core._ColumnOptions.SEARCH
        layers_model = LayerTableModel(columns=self._LAYERS_COLUMNS)
        self._layers = _LayersSheet(layers_model, self._LAYERS_COLUMNS, options)
        self._prims = _sheets._StageSpreadsheet(columns=self._PRIM_COLUMNS, options=options)
        # TODO: better APIs for this default setup?
        self._prims.sorting_enabled.setVisible(False)
        self._prims._model_hierarchy.setChecked(False)
        for each in (self._prims._classes, self._prims._orphaned, self._prims._inactive):
            each.setChecked(True)
        self._prims._filters_logical_op.setCurrentIndex(1)
        for each in self._layers, self._prims:
            each.layout().setContentsMargins(0,0,0,0)

        horizontal = QtWidgets.QSplitter(QtCore.Qt.Horizontal)
        horizontal.addWidget(self._layers)
        horizontal.addWidget(self._prims)

        self._graph_view = _graph._GraphViewer(parent=self)
        self.setFocusProxy(self._graph_view)
        _graph_legend_controls = QtWidgets.QFrame()
        _graph_controls_layout = QtWidgets.QVBoxLayout()
        _graph_arcs_layout = QtWidgets.QHBoxLayout()

        _graph_controls_layout.addLayout(_graph_arcs_layout)
        _graph_legend_controls.setLayout(_graph_controls_layout)
        self._graph_edge_include = _graph_edge_include = {}
        self._graph_precise_source_ports = QtWidgets.QCheckBox("Precise Source Layer")
        self._graph_precise_source_ports.setChecked(False)
        self._graph_precise_source_ports.clicked.connect(lambda: self._update_graph_from_graph_info(self._computed_graph_info))
        for arc, arc_details in _ARCS_LEGEND.items():
            arc_btn = QtWidgets.QCheckBox(arc.displayName.title())
            arc_btn.setStyleSheet(f"background-color: {arc_details['color']}; padding: 3px; border-width: 1px; border-radius: 3;")
            arc_btn.setChecked(True)
            _graph_arcs_layout.addWidget(arc_btn)
            _graph_edge_include[arc] = arc_btn
            arc_btn.clicked.connect(lambda: self._update_graph_from_graph_info(self._computed_graph_info))
        _graph_arcs_layout.addWidget(self._graph_precise_source_ports)

        filters_layout = QtWidgets.QHBoxLayout()
        _graph_controls_layout.addLayout(filters_layout)
        # Arcs filters
        def _arc_filter(title, state=QtCore.Qt.CheckState.PartiallyChecked):
            widget = QtWidgets.QCheckBox(title)
            widget.setTristate(True)
            widget.setCheckState(state)
            widget.stateChanged.connect(self._edge_filter_changed)
            # widget.toggled
            filters_layout.addWidget(widget)
            return widget

        _graph_arcs_layout.addStretch(0)
        # self._has_specs = _arc_filter("Has Specs", QtCore.Qt.CheckState.Checked)
        self._has_specs = _arc_filter("Has Specs")
        self._is_ancestral = _arc_filter("Is Ancestral")
        self._is_implicit = _arc_filter("Is Implicit")
        self._from_root_prim_spec = _arc_filter("From Root Layer Prim Spec")
        self._from_root_layer_stack = _arc_filter("From Root LayerStack")
        filters_layout.addStretch(0)
        ##############

        _graph_legend_controls.setFixedHeight(_graph_legend_controls.sizeHint().height())
        vertical = QtWidgets.QSplitter(QtCore.Qt.Vertical)
        vertical.addWidget(horizontal)
        vertical.addWidget(_graph_legend_controls)
        vertical.addWidget(self._graph_view)
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(vertical)
        self.setLayout(layout)
        selectionModel = self._layers.table.selectionModel()
        selectionModel.selectionChanged.connect(self._selectionChanged)
        self._prim_paths_to_compute = set()
        self.setWindowTitle("LayerStack Composition")

    def _edge_filter_changed(self, *args, **kwargs):
        state = lambda widget: None if widget.checkState() == QtCore.Qt.CheckState.PartiallyChecked else widget.isChecked()
        edge_data_filter = _nx_graph_edge_filter(
            has_specs=state(self._has_specs),
            ancestral=state(self._is_ancestral),
            implicit=state(self._is_implicit),
            introduced_in_root_layer_prim_spec=state(self._from_root_prim_spec),
            introduced_in_root_layer_stack=state(self._from_root_layer_stack),
        )
        graph = self._graph_view.graph
        # self._graph_view.filter_edges = (lambda *edge: edge_data_filter(graph.edges[edge])) if edge_data_filter else None
        from pprint import pp
        def actual(*edge):
            # edge_info = graph.edges[edge]
            # pp(edge_info)
            # print(f"{edge_data_filter=}")
            result = edge_data_filter(graph.edges[edge])
            # print(f"{result=}")
            return result

        self._graph_view.filter_edges = actual if edge_data_filter else None
        if isinstance(self._graph_view, _graph._GraphSVGViewer):
            self._graph_view._subgraph_dot_path.cache_clear()
        self._graph_view.view(self._graph_view._viewing)

    def _selectionChanged(self, selected: QtCore.QItemSelection, deselected: QtCore.QItemSelection):
        node_ids = {index.data(_core._QT_OBJECT_DATA_ROLE) for index in self._layers.table.selectedIndexes()}
        node_indices = set(chain.from_iterable(self._computed_graph_info.ids_by_layers[layer] for layer in node_ids))

        prims_model = self._prims.model
        prims_model._root_paths = paths = set(chain.from_iterable(
            # some layers from the layer stack might be on our selected indices but they wont be on the paths_by_ids
            self._computed_graph_info.paths_by_ids[i] for i in node_indices if i in self._computed_graph_info.paths_by_ids)
        )
        all_paths = set(chain.from_iterable(self._computed_graph_info.paths_by_ids.values()))
        self._prims._filter_predicate = lambda prim: prim.GetPath() in (paths or all_paths)
        self._prims._update_stage()
        self._graph_view.view(node_indices)

    @_core.wait()
    def setStage(self, stage):
        """Sets the USD stage the spreadsheet is looking at."""
        self._stage = stage
        self._layers._resolver_context = stage.GetPathResolverContext()
        predicate = Usd.TraverseInstanceProxies(Usd.PrimAllPrimsPredicate)
        prims = Usd.PrimRange.Stage(stage, predicate)
        # prims = (prim for prim in prims if prim.IsActive())
        if self._prim_paths_to_compute:
            prims = (p for p in prims if p.GetPath() in self._prim_paths_to_compute)
            self._prims._filter_predicate = lambda prim: prim.GetPath() in self._prim_paths_to_compute

        graph_info = _compute_layerstack_graph(prims, self._graph_view.url_id_prefix)
        self._prims.setStage(stage)
        # self._edge_filter_changed()
        self._update_graph_from_graph_info(graph_info)

    def setPrimPaths(self, value):
        self._prim_paths_to_compute = {p if isinstance(p, Sdf.Path) else Sdf.Path(p) for p in value}

    def _update_graph_from_graph_info(self, graph_info: _GraphInfo):
        self._computed_graph_info = graph_info
        # https://stackoverflow.com/questions/33262913/networkx-move-edges-in-nx-multidigraph-plot
        graph = nx.MultiDiGraph()
        graph.graph['graph'] = dict(tooltip="LayerStack Composition")
        graph.add_nodes_from(self._computed_graph_info.nodes.items())
        graph.add_edges_from(self._iedges(graph_info))
        self._graph_view._graph = graph
        self._graph_view.sticky_nodes.extend(graph_info.sticky_nodes)
        self._layers.model.setLayers(graph_info.ids_by_layers)
        # view intersection as we might be seeing nodes that no longer exist
        # self._graph_view.view(self._graph_view._viewing)  # TODO: check if this is better <- and reset _viewing somewhere else
        ## NEW
        self._graph_view._viewing = self._graph_view._viewing.intersection(graph_info.nodes)
        self._edge_filter_changed()
        ## NEW END
        # self._graph_view.view(self._graph_view._viewing.intersection(graph_info.nodes))

    def _iedges(self, graph_info: _GraphInfo):
        checked_arcs = {arc for arc, control in self._graph_edge_include.items() if control.isChecked()}
        precise_ports = self._graph_precise_source_ports.isChecked()
        for (src, tgt), edge_info in graph_info.edges.items():
            arc_ports = edge_info if precise_ports else {None: collections.ChainMap(*edge_info.values())}
            for src_port, arcs in arc_ports.items():
                visible_arcs = {arc: attrs for arc, attrs in arcs.items() if (arc in checked_arcs or {src, tgt}.issubset(graph_info.sticky_nodes))}
                if visible_arcs:
                    # Composition arcs target layer stacks, so we don't specify port on our target nodes
                    ports = {"tailport": src_port} if precise_ports and src_port is not None else {}
                    color = _edge_color(tuple(visible_arcs))
                    yield src, tgt, collections.ChainMap(ports, color, *visible_arcs.values())

